# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Optional, Tuple

import numpy as np
import torch

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.definitions import NNCFGraphNodeType
from nncf.common.graph.transformations.commands import TargetType
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.quantization.algorithms.fast_bias_correction.backend import FastBiasCorrectionAlgoBackend
from nncf.tensor import Tensor
from nncf.torch.graph.transformations.command_creation import create_bias_correction_command
from nncf.torch.graph.transformations.commands import PTBiasCorrectionCommand
from nncf.torch.graph.transformations.commands import PTModelExtractionCommand
from nncf.torch.graph.transformations.commands import PTTargetPoint
from nncf.torch.model_graph_manager import get_fused_bias_value
from nncf.torch.model_graph_manager import get_potential_fused_node
from nncf.torch.model_graph_manager import is_node_with_fused_bias
from nncf.torch.model_graph_manager import is_quantized_weights
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.tensor_statistics.collectors import get_mean_statistic_collector


class PTFastBiasCorrectionAlgoBackend(FastBiasCorrectionAlgoBackend):
    TARGET_TYPE_TO_PT_INS_TYPE_MAP = {
        TargetType.PRE_LAYER_OPERATION: TargetType.OPERATOR_PRE_HOOK,
        TargetType.POST_LAYER_OPERATION: TargetType.OPERATOR_POST_HOOK,
    }

    @staticmethod
    def target_point(target_type: TargetType, target_node_name: str, port_id: int) -> PTTargetPoint:
        if NNCFGraphNodeType.INPUT_NODE in target_node_name or target_type == TargetType.POST_LAYER_OPERATION:
            port_id = None
        if target_type in PTFastBiasCorrectionAlgoBackend.TARGET_TYPE_TO_PT_INS_TYPE_MAP:
            target_type = PTFastBiasCorrectionAlgoBackend.TARGET_TYPE_TO_PT_INS_TYPE_MAP[target_type]
        return PTTargetPoint(target_type, target_node_name, input_port_id=port_id)

    @staticmethod
    def create_bias_correction_command(
        node: NNCFNode, bias_value: Tensor, nncf_graph: NNCFGraph
    ) -> PTBiasCorrectionCommand:
        return create_bias_correction_command(node, bias_value.data)

    @staticmethod
    def model_extraction_command(
        input_ids: List[Tuple[str, int]], output_ids: List[Tuple[str, int]]
    ) -> PTModelExtractionCommand:
        return PTModelExtractionCommand([input_ids[0][0]], [output_ids[0][0]])

    @staticmethod
    def mean_statistic_collector(
        channel_axis: int,
        inplace: bool,
        num_samples: Optional[int] = None,
        window_size: Optional[int] = None,
    ) -> TensorCollector:
        return get_mean_statistic_collector(num_samples, channel_axis, window_size)

    @staticmethod
    def get_sub_input_output_names(subgraph: NNCFNetwork) -> Tuple[Optional[str], Optional[str]]:
        # Pytorch does not have name for extracted node
        return None, None

    @staticmethod
    def create_input_data(shape: Tuple[int], data: List[Tensor], input_name: str, channel_axis: int) -> torch.Tensor:
        blob = torch.zeros(shape, dtype=data[0].data.dtype, device=data[0].data.device)
        for j, idx in enumerate(np.ndindex(blob.shape[channel_axis])):
            index = tuple(slice(None) if i != channel_axis else idx for i in range(blob.ndim))
            blob[index] = data[j].data
        return blob

    @staticmethod
    def get_bias_value(node: NNCFNode, nncf_graph: NNCFGraph, model: NNCFNetwork) -> Tensor:
        return Tensor(get_fused_bias_value(node, model))

    @staticmethod
    def get_activation_port_ids_for_bias_node(node: NNCFNode) -> Tuple[int, int]:
        return 0, 0

    @staticmethod
    def process_model_output(raw_data: Dict, output_name: Optional[str]) -> Tensor:
        return Tensor(raw_data)

    @staticmethod
    def is_quantized_weights(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_quantized_weights(node, nncf_graph)

    @staticmethod
    def is_node_with_bias(node: NNCFNode, nncf_graph: NNCFGraph) -> bool:
        return is_node_with_fused_bias(node, nncf_graph)

    @staticmethod
    def get_node_names_for_input_output_statistics(node: NNCFNode, nncf_graph: NNCFGraph) -> Tuple[str, str]:
        input_node_name = node.node_name
        next_norm_node = get_potential_fused_node(input_node_name, nncf_graph)
        output_node_name = next_norm_node.node_name if next_norm_node else input_node_name
        return input_node_name, output_node_name

    @staticmethod
    def get_activation_channel_axis(node: NNCFNode, pord_id: int, input_shape: Tuple[int]) -> int:
        return node.metatype.output_channel_axis



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Counter
from collections import defaultdict
from copy import deepcopy
from typing import Dict, List, Optional, Tuple, TypeVar

import nncf
from nncf import Dataset
from nncf.common.factory import ModelTransformerFactory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.graph import NNCFNode
from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.graph.utils import get_reduction_axes
from nncf.common.logging import nncf_logger
from nncf.common.logging.track_progress import track
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.quantization.algorithms.algorithm import Algorithm
from nncf.tensor import Tensor
from nncf.tensor import functions as fns

TModel = TypeVar("TModel")
TTensor = TypeVar("TTensor")
STATISTIC_BRANCH_KEY = "abs_max"
ALPHA_MAP = {"convolution": 0.05, "matmul": 0.95}


class SmoothQuant(Algorithm):
    """
    Post-training SmoothQuant algorithm implementation.

    The main purpose of this algorithm is to reduce activation quantization error
    via the insertion of nodes with smoothing scales for weighted layers.
    """

    def __init__(
        self, subset_size: int = 300, inplace_statistics: bool = True, alpha_map: Dict[str, float] = ALPHA_MAP
    ):
        """
        :param subset_size: Size of a subset for the statistics collection,
            default is 300.
        :param inplace_statistics: Defines whether to calculate quantizers statistics
            by backend graph operations or by default Python implementation, defaults
            to True.
        :param alpha_map: The parameter that regulates the calculation of the scale for different layers.
            The default value for each layer in the ALPHA_MAP.
            Negative value switches off the algorithm for correspondent nodes.
        """

        super().__init__()
        self._subset_size = subset_size
        self._inplace_statistics = inplace_statistics
        self._backend_entity = None
        self._algorithm_key = f"SQ_{hash(self)}"
        self._cached_multiply_names = Counter()
        self._alpha_map = alpha_map

    @property
    def available_backends(self) -> List[BackendType]:
        return [BackendType.OPENVINO, BackendType.TORCH, BackendType.TORCH_FX]

    def _set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backed-specific logic of the algorithm.

        :param model: Backend-specific input model.
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.smooth_quant.openvino_backend import OVSmoothQuantAlgoBackend

            self._backend_entity = OVSmoothQuantAlgoBackend()
        elif model_backend == BackendType.TORCH:
            from nncf.quantization.algorithms.smooth_quant.torch_backend import PTSmoothQuantAlgoBackend

            self._backend_entity = PTSmoothQuantAlgoBackend()
        elif model_backend == BackendType.TORCH_FX:
            from nncf.quantization.algorithms.smooth_quant.torch_fx_backend import FXSmoothQuantAlgoBackend

            self._backend_entity = FXSmoothQuantAlgoBackend()
        else:
            raise nncf.UnsupportedBackendError(
                "Cannot return backend-specific entity because {} is not supported!".format(model_backend.value)
            )

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> TModel:
        self._set_backend_entity(model)
        alpha_map = self._get_alpha_map()

        nodes_to_smooth_data = self._get_nodes_to_smooth_data(graph, alpha_map.keys())
        model_transformer = ModelTransformerFactory.create(model)
        transformation_layout = TransformationLayout()

        node_groups = self._group_nodes_by_source(nodes_to_smooth_data, graph)

        for group_id, nodes in track(node_groups.items(), description="Applying Smooth Quant"):
            best_scale = None
            best_ratio = 0.0
            empty_statistic = False
            for node_to_smooth in nodes:
                source_node, input_port_id, source_output_port_id, _ = group_id
                activations_value = self._get_statistics_for_node(
                    statistic_points, node_to_smooth.node_name, input_port_id
                )
                if any(val is None for val in activations_value):
                    empty_statistic = True
                    break
                if len(activations_value) != 1:
                    raise RuntimeError(
                        (
                            "More than one statistic is collected for one node during"
                            f"Smooth Quanti algorithm: {node_to_smooth.node_name}"
                        )
                    )

                activations_value = self._clip_statistics(activations_value)

                weight_value = self._backend_entity.get_weight_value(node_to_smooth, model, graph)
                weight_statistics = self._process_weight_statistics(node_to_smooth, weight_value)
                weight_statistics = self._clip_statistics([weight_statistics])

                alpha = alpha_map[node_to_smooth.metatype]

                scales, ratio = self._calculate_scale_and_ratio(activations_value, weight_statistics, alpha)

                if ratio > best_ratio:
                    best_ratio = ratio
                    best_scale = deepcopy(scales)

            if empty_statistic:
                nncf_logger.debug(
                    f"Skipped SmoothQuant for nodes after {source_node.node_name} because of the empty statistics."
                )
                continue

            if best_scale is None:
                nncf_logger.debug(
                    f"Skipped SmoothQuant for nodes after {source_node.node_name} because of the empty scale."
                )
                continue

            for node_to_smooth in nodes:
                weight_value = self._backend_entity.get_weight_value(node_to_smooth, model, graph)
                weights_scale = self._calculate_weight_scale(best_scale, node_to_smooth, weight_value)
                scaled_weight = weight_value * weights_scale
                weight_update_command = self._backend_entity.weight_update_command(node_to_smooth, scaled_weight.data)
                transformation_layout.register(weight_update_command)

            activations_by_output_id = {e.output_port_id: e for e in graph.get_output_edges(source_node)}
            activations_shape = activations_by_output_id[source_output_port_id].tensor_shape
            activation_scale = self._calculate_activation_scale(best_scale, activations_shape, nodes, graph)

            scale_node_name = self._create_scale_node_name(source_node.node_name, source_output_port_id)
            scale_insertion_command = self._backend_entity.scale_insertion_command(
                source_node, activation_scale.data, source_output_port_id, nodes, scale_node_name
            )
            transformation_layout.register(scale_insertion_command)

        transformed_model = model_transformer.transform(transformation_layout)
        return transformed_model

    @staticmethod
    def _calculate_scale_and_ratio(
        activations: Tensor, weights: Tensor, alpha: float, quantile: Optional[float] = 0.1
    ) -> Tuple[Tensor, float]:
        """
        Calculates base scale value and it's ratio.

        :param activations: Activation statistics value.
        :param weights: Weights statistics value.
        :param alpha: Base value for exponentiation.
        :param quantile: Base quantile value.
        :return: Calculated base scale value & ratio.
        """

        eps = fns.finfo(activations).eps
        scales = fns.power(activations, alpha) / (fns.power(weights, 1 - alpha) + eps)

        a_min = fns.quantile(scales, quantile, keepdims=False)
        a_max = 1e2

        scales = fns.clip(scales, a_min=a_min, a_max=a_max)
        ratio = scales.min() / (scales.max() + eps)
        return scales, ratio

    def _group_nodes_by_source(self, nodes_to_smooth: List[Dict], nncf_graph: NNCFGraph) -> Dict[tuple, List]:
        """
        Groups nodes that will be smoothed by source (parent node).

        :param nodes_to_smooth: List of the nodes that will be smoothed.
        :param nncf_graph: NNCFGraph instance.
        :return: Dictionary with the source info as key and grouped nodes as value.
        """
        groups = defaultdict(list)
        for node_data in nodes_to_smooth:
            node_to_smooth = node_data["node_to_smooth"]
            input_act_port = node_data["input_act_port"]
            source_node = nncf_graph.get_input_edge_by_port_id(node_to_smooth, input_act_port).from_node
            edge = nncf_graph.get_edge(source_node, node_to_smooth)
            # Such group_id (with node, ports, and shape as a hash) allows us to be confident
            # that all sensitive parameters are equal for successor nodes are equal.
            group_id = (source_node, input_act_port, edge.output_port_id, hash(str(edge.tensor_shape)))
            groups[group_id].append(node_to_smooth)

        return groups

    def _get_statistics_for_node(
        self, statistic_points: StatisticPointsContainer, node_name: str, act_port: int
    ) -> List[TTensor]:
        """
        Collects statistics for node.

        :param statistic_points: StatisticPointsContainer instance.
        :param node_name: Name of the node for collection.
        :param act_port: Activation port id.
        :return: List of the TTensor instances.
        """

        statistics_for_node = []
        for tensor_collector in statistic_points.get_algo_statistics_for_node(
            node_name,
            self._backend_entity.get_filter_fn_for_statistics(act_port, self._algorithm_key),
            self._algorithm_key,
        ):
            statistic = tensor_collector.get_statistics()[STATISTIC_BRANCH_KEY]
            statistics_for_node.append(statistic)
        return statistics_for_node

    def get_statistic_points(self, model: TModel, graph: NNCFGraph) -> StatisticPointsContainer:
        statistic_container = StatisticPointsContainer()

        self._set_backend_entity(model)
        alpha_map = self._get_alpha_map()

        nodes_to_smooth_data = self._get_nodes_to_smooth_data(graph, alpha_map.keys())

        for node_data in nodes_to_smooth_data:
            node_to_smooth = node_data["node_to_smooth"]
            target_point = self._backend_entity.target_point(
                target_type=self._backend_entity.pre_layer_target_type(),
                target_node_name=node_to_smooth.node_name,
                port_id=node_data["input_act_port"],
            )
            input_reduction_axes = self._calculate_input_reduction_axes(
                graph, node_to_smooth, node_data["input_act_port"]
            )
            stat_collector = self._backend_entity.get_abs_max_channel_collector(
                self._subset_size, input_reduction_axes, self._inplace_statistics, STATISTIC_BRANCH_KEY
            )
            statistic_container.add_statistic_point(
                StatisticPoint(
                    target_point=target_point,
                    tensor_collector=stat_collector,
                    algorithm=self._algorithm_key,
                )
            )
        return statistic_container

    def _get_nodes_to_smooth_data(self, nncf_graph: NNCFGraph, node_metatypes: List[OperatorMetatype]) -> List[Dict]:
        """
        Collects layers whose activations will be smoothed.

        :param nncf_graph: NNCFGraph instance.
        :param node_metatypes: Metatypes for nodes to search for.
        :return: List with the data for each layer.
        """
        nodes_with_weights = nncf_graph.get_nodes_by_metatypes(node_metatypes)
        nodes_to_smooth_data = []

        for node_with_weight in nodes_with_weights:
            if not self._backend_entity.is_node_with_weights(node_with_weight):
                continue

            activation_port_id = self._backend_entity.get_activations_port_id(node_with_weight, nncf_graph)
            activation_node = nncf_graph.get_input_edge_by_port_id(node_with_weight, activation_port_id).from_node

            # Skipping agnostic layers as inputs to propagate quantizer
            # Only for Convolution layers
            if (
                node_with_weight.metatype in self._backend_entity.convolution_metatypes
                and activation_node.metatype in self._backend_entity.quantize_agnostic_metatypes
            ):
                continue

            # Skipping shared weights
            if self._backend_entity.is_node_with_shared_weight(node_with_weight, nncf_graph):
                continue

            nodes_to_smooth_data.append(
                {
                    "node_to_smooth": node_with_weight,
                    "input_act_port": activation_port_id,
                }
            )
        return nodes_to_smooth_data

    def _calculate_activation_scale(
        self, scale_value: TTensor, activations_shape: List[int], nodes: List[NNCFNode], nncf_graph: NNCFGraph
    ) -> TTensor:
        """
        Calculates activation scales for Smooth node.

        :param scale_value: Base scale value.
        :param activations_shape: activation tensor shape.
        :param nodes: List of consumers for Smooth node.
        :return: Calculated per-channel activation scale.
        """
        activation_ports_map = {node: self._backend_entity.get_activations_port_id(node, nncf_graph) for node in nodes}
        channel_axes = [
            self._backend_entity.get_activation_channel_axis(node, port) for node, port in activation_ports_map.items()
        ]
        channel_axis = channel_axes[0]

        if not all(axis == channel_axis for axis in channel_axes):
            raise nncf.InternalError(f"Channel axes for nodes {[n.node_name for n in nodes]} are not identical")

        activations_size = len(activations_shape)
        activation_scale = scale_value ** (-1)
        if activations_size > 1:
            reshape_shape = [1 for _ in range(activations_size)]
            reshape_shape[channel_axis] = activation_scale.size
            activation_scale = activation_scale.reshape(reshape_shape)
        return activation_scale

    def _calculate_weight_scale(self, scale_value: Tensor, node: NNCFNode, weights_value: Tensor) -> Tensor:
        """
        Calculates scale for weight tensor.

        :param scale_value: Base scale value.
        :param node: Consumer for Smooth node.
        :return: Calculated scale for weights.
        """
        weights_size = len(weights_value.shape)
        if weights_size > 1:
            channel_axis = self._backend_entity.get_weight_channel_axis(node)
            weight_scale = scale_value
            if weights_size > 1:
                reshape_shape = [1 for _ in range(weights_size)]
                reshape_shape[channel_axis] = scale_value.size
                weight_scale = scale_value.reshape(reshape_shape)
            return weight_scale
        return scale_value

    def _calculate_input_reduction_axes(self, nncf_graph: NNCFGraph, node: NNCFNode, input_port: int) -> Tuple[int]:
        """
        Returns reduction axes for specified input.

        :param nncf_graph: NNCFGraph instance.
        :param node: NNCFNode to check.
        :param input_port: Specified input port id.
        :return: Calculated reduction axes.
        """
        shape = nncf_graph.get_input_edge_by_port_id(node, input_port).tensor_shape
        reduction_axes = tuple([])
        if len(shape) > 1:
            channel_axis = self._backend_entity.get_activation_channel_axis(node, input_port)
            reduction_axes = get_reduction_axes((channel_axis,), shape)
        return reduction_axes

    def _process_weight_statistics(self, node: NNCFNode, weights: Tensor) -> Tensor:
        """
        Returns processed weight statistics for node.

        :param node: NNCFNode to check.
        :param weights: Backend-specific weights.
        :return: Weight statistic for node.
        """
        channel_axis = 0
        if len(weights.shape) > 1:
            channel_axis = self._backend_entity.get_weight_channel_axis(node)
        reduction_shape = [i for i, _ in enumerate(weights.shape)]
        reduction_shape.pop(channel_axis)
        return fns.max(fns.abs(weights), axis=tuple(reduction_shape))

    def _create_scale_node_name(self, source_name: str, source_port_id: int) -> str:
        """
        Returns uniqie scale node name for new layer.

        :param source_name: Source layer name.
        :param source_port_id: Source port id.
        :return: Generated uniqie name.
        """
        scale_node_name = f"{source_name}_{source_port_id}"
        unique_index = self._cached_multiply_names[scale_node_name]
        self._cached_multiply_names[scale_node_name] += 1
        return f"{scale_node_name}_{unique_index}/nncf_smooth_quant"

    def _get_alpha_map(self) -> Dict[OperatorMetatype, float]:
        """
        Returns alpha map by metatypes.

        :return: Alpha map by metatypes.
        """
        alpha_by_metatype_map = {}
        name_to_metatype = {
            "convolution": self._backend_entity.convolution_metatypes,
            "matmul": self._backend_entity.matmul_metatypes,
        }
        for type_name, alpha_value in self._alpha_map.items():
            if alpha_value < 0:
                nncf_logger.debug(
                    f"Smooth Quant algorithm does not support negative parameter for {type_name}! "
                    "Skipping these layers."
                )
                continue
            metatypes = name_to_metatype[type_name]
            for metatype in metatypes:
                alpha_by_metatype_map[metatype] = alpha_value
        return alpha_by_metatype_map

    @staticmethod
    def _clip_statistics(statistics: List[Tensor]) -> Tensor:
        """
        Clips statistics for further calculation.
        :param statistics: Input statistics.
        :return: Clipped statistics.
        """
        a_min = 1e-5

        statistics = fns.stack(statistics)
        squeezed = fns.squeeze(statistics)
        return fns.clip(squeezed, a_min=a_min, a_max=None)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
from copy import deepcopy
from functools import partial
from inspect import Parameter
from inspect import Signature
from typing import Any, Dict, List, Optional, Protocol, Set, Tuple, Type, Union

import torch

import nncf
from nncf import Dataset
from nncf import NNCFConfig
from nncf.common.graph.definitions import MODEL_INPUT_OP_NAME
from nncf.common.graph.definitions import MODEL_OUTPUT_OP_NAME
from nncf.common.initialization.dataloader import NNCFDataLoader
from nncf.common.logging import nncf_logger
from nncf.common.utils.api_marker import api
from nncf.config.structures import BNAdaptationInitArgs
from nncf.config.structures import QuantizationRangeInitArgs
from nncf.torch.dynamic_graph.context import forward_nncf_trace
from nncf.torch.dynamic_graph.patch_pytorch import register_operator
from nncf.torch.initialization import wrap_dataloader_for_init
from nncf.torch.nested_objects_traversal import objwalk
from nncf.torch.utils import get_model_device
from nncf.torch.utils import is_tensor
from nncf.torch.utils import is_traced_tensor


@api(canonical_alias="nncf.torch.nncf_model_input")
@register_operator(name=MODEL_INPUT_OP_NAME)
def nncf_model_input(tensor: "torch.Tensor"):
    return tensor


@api(canonical_alias="nncf.torch.nncf_model_output")
@register_operator(name=MODEL_OUTPUT_OP_NAME)
def nncf_model_output(tensor: "torch.Tensor"):
    return tensor


def wrap_nncf_model_inputs_with_objwalk(model_args, model_kwargs):
    model_args = objwalk(model_args, is_tensor, nncf_model_input)
    model_kwargs = objwalk(model_kwargs, is_tensor, nncf_model_input)
    return model_args, model_kwargs


def wrap_nncf_model_outputs_with_objwalk(model_outputs):
    model_outputs = objwalk(model_outputs, is_traced_tensor, nncf_model_output)
    return model_outputs


def replicate_same_tensors(obj: Any) -> Any:
    """
    Required to handle the situation when multiple references to one and the
    same tensor are present in the input. If tensor replication is not done, then
    at runtime one and the same tensor could be wrapped by input/output wrappers twice,
    which will disrupt the traced graph structure and possibly hook calls.
    """
    observed_tensor_object_ids: Set[int] = set()

    def replicate_fn(tensor: torch.Tensor) -> torch.Tensor:
        tensor_object_id = id(tensor)
        if tensor_object_id in observed_tensor_object_ids:
            with forward_nncf_trace():
                return tensor.clone()
        observed_tensor_object_ids.add(tensor_object_id)
        return tensor

    obj = objwalk(obj, is_tensor, replicate_fn)
    return obj


class ModelInputInfo(abc.ABC):
    """
    Provides the model inputs that are subsequently used as example inputs to the forward function of
    the compressed model for purposes of compression (e.g. at the stage of building the dynamic graph, exporting
    to a serialized state etc.)
    """

    @abc.abstractmethod
    def get_forward_inputs(self, device: Optional[Union[str, torch.device]] = None) -> Tuple[Tuple, Dict]:
        """
        Returns the tuple of (args, kwargs) for passing into the compressed model's forward method when necessary.
        The returned arguments should be such that the model's forward with these arguments executes the main
        control flow of the model that needs to be compressed.

        :param device: Optional - a PyTorch string representation of the device that the tensors among the returned
          args, kwargs should be located at.
        """
        pass


class FillerInputElement:
    """
    Represents a single tensor argument (positional or keyword) in the model's example input that is to be generated
    on the fly and filled with a requested type of data filler.
    """

    FILLER_TYPE_ONES = "ones"
    FILLER_TYPE_ZEROS = "zeros"
    FILLER_TYPE_RANDOM = "random"
    FILLER_TYPES = [FILLER_TYPE_ONES, FILLER_TYPE_ZEROS, FILLER_TYPE_RANDOM]

    def __init__(self, shape: List[int], type_str: str = "float", keyword: str = None, filler: str = None):
        """
        :param shape: The shape of the model input tensor.
        :param type_str: The type of the model input tensor - "float" for torch.float32, "long" for torch.long
        :param keyword: Optional - if specified, then this input tensor will be passed as a corresponding keyword
          parameter, and as a positional argument if this parameter is unspecified.
        :param filler: Optional - can be either "ones", "zeros" or "random". The model input tensor will be generated
          with data corresponding to this setting. Default is "ones".
        """
        self.shape = shape
        self.type = self._string_to_torch_type(type_str)
        self.keyword = keyword
        if filler is None:
            self.filler = self.FILLER_TYPE_ONES
        else:
            self.filler = filler
            if self.filler not in self.FILLER_TYPES:
                raise ValueError(f"Unknown input filler type: {filler}")

    @staticmethod
    def _string_to_torch_type(string):
        if string == "long":
            return torch.long
        return torch.float32

    @staticmethod
    def torch_type_to_string(dtype: torch.dtype):
        if dtype is torch.long:
            return "long"
        return "float"

    def is_integer_input(self):
        return self.type != torch.float32

    def __eq__(self, other: "FillerInputElement"):
        return self.type == other.type and self.keyword == other.keyword and self.filler == other.filler

    def get_tensor_for_input(self) -> torch.Tensor:
        if self.filler == FillerInputElement.FILLER_TYPE_ZEROS:
            return torch.zeros(size=self.shape, dtype=self.type)
        if self.filler == FillerInputElement.FILLER_TYPE_ONES:
            return torch.ones(size=self.shape, dtype=self.type)
        if self.filler == FillerInputElement.FILLER_TYPE_RANDOM:
            return torch.rand(size=self.shape, dtype=self.type)
        raise NotImplementedError


class FillerInputInfo(ModelInputInfo):
    """
    An implementation of ModelInputInfo that defines the model input in terms of shapes and types of individual
    tensor args and kwargs of the model's forward method.
    """

    def __init__(self, elements: List[FillerInputElement]):
        super().__init__()
        self.elements = deepcopy(elements)

    @classmethod
    def from_nncf_config(cls, config: NNCFConfig):
        """
        Parses the NNCFConfig's "input_info" field if it is present to determine model input information,
        otherwise raises a RuntimeError. The "input_info" field structure must conform to the NNCF config jsonschema.

        :param config: An NNCFConfig instance.
        :return: FillerInputInfo object initialized according to config.
        """
        input_infos = config.get("input_info")
        if input_infos is None:
            raise nncf.ValidationError("Passed NNCFConfig does not have an 'input_info' field")
        if isinstance(input_infos, dict):
            return FillerInputInfo(
                [
                    FillerInputElement(
                        input_infos.get("sample_size"),
                        input_infos.get("type"),
                        input_infos.get("keyword"),
                        input_infos.get("filler"),
                    )
                ]
            )
        if isinstance(input_infos, list):
            elements: List[FillerInputElement] = []
            for info_dict in input_infos:
                elements.append(
                    FillerInputElement(
                        info_dict.get("sample_size"),
                        info_dict.get("type"),
                        info_dict.get("keyword"),
                        info_dict.get("filler"),
                    )
                )
            return FillerInputInfo(elements)
        raise nncf.ValidationError("Invalid input_infos specified in config - should be either dict or list of dicts")

    def get_forward_inputs(
        self, device: Optional[Union[str, torch.device]] = None
    ) -> Tuple[Tuple[torch.Tensor, ...], Dict[str, torch.Tensor]]:
        args_list = []
        kwargs = {}
        for fe in self.elements:
            tensor = fe.get_tensor_for_input()
            if device is not None:
                tensor = tensor.to(device)
            if fe.keyword is None:
                args_list.append(tensor)
            else:
                kwargs[fe.keyword] = tensor
        return tuple(args_list), kwargs


class ExactInputsInfo(ModelInputInfo):
    """
    An implementation of ModelInputInfo that defines the model input in terms of exact forward args and kwargs.
    """

    def __init__(self, forward_args: Tuple, forward_kwargs: Dict):
        self._forward_args = forward_args
        self._forward_kwargs = forward_kwargs

    def get_forward_inputs(self, device: Optional[Union[str, torch.device]] = None) -> Tuple[Tuple, Dict]:
        if device is None:
            return self._forward_args, self._forward_kwargs
        to_device_fn = partial(torch.Tensor.to, device=device)
        args_copy = deepcopy(self._forward_args)
        kwargs_copy = deepcopy(self._forward_kwargs)
        args_at_device = objwalk(args_copy, is_tensor, to_device_fn)
        kwargs_at_device = objwalk(kwargs_copy, is_tensor, to_device_fn)
        return args_at_device, kwargs_at_device


class ExampleInputInfo(ExactInputsInfo):
    @classmethod
    def from_example_input(cls, example_input: Any) -> "ExampleInputInfo":
        """
        Builds an ExampleInputInfo object based on the example input.

        :param dataset: An nncf.Dataset whose first element will be used as an example model input
        :return: An initialized ExampleInputInfo object.
        """
        if isinstance(example_input, tuple):
            return ExampleInputInfo(example_input, {})
        if isinstance(example_input, dict):
            return ExampleInputInfo(tuple(), example_input)
        return ExampleInputInfo((example_input,), {})

    @classmethod
    def from_nncf_dataset(cls, dataset: Dataset) -> "ExampleInputInfo":
        """
        Checks the first element of the provided nncf.Dataset and builds an ExampleInputInfo object that would
        provide the same input to the model at corresponding compression stages.

        :param dataset: An nncf.Dataset whose first element will be used as an example model input
        :return: An initialized ExampleInputInfo object.
        """
        example_input = next(iter(dataset.get_inference_data()))
        return cls.from_example_input(example_input)


class LoaderInputInfo(ExactInputsInfo):
    @classmethod
    def from_nncf_config_dataloaders(cls, config: NNCFConfig) -> Optional["LoaderInputInfo"]:
        """
        Examines the user-provided structures registered with the NNCFConfig instance used for compression to find
        structures that contain a dataloader. The dataloader's first element is used to provide an example input to
        the compressed model.

        :param config: An nncf.NNCFConfig instance. Must have at least one NNCFExtraConfigStruct attached that can
          provide a dataloader (these are listed in nncf.torch.dynamic_graph.io_handling.EXTRA_STRUCTS_WITH_DATALOADERS)

        :return: An initialized LoaderInputInfo object.
        """
        extra_structs = config.get_all_extra_structs()
        for extra_struct in extra_structs:
            if isinstance(extra_struct, tuple(EXTRA_STRUCTS_WITH_DATALOADERS)):
                extra_struct: HasDataloader
                dataloader = extra_struct.data_loader
                wrapped_dataloader = wrap_dataloader_for_init(dataloader)
                dataloader_output = next(iter(wrapped_dataloader))
                args, kwargs = wrapped_dataloader.get_inputs(dataloader_output)
                return LoaderInputInfo(args, kwargs)
        # config extra structs had no suitable dataloaders
        return None


class InputInfoWrapManager:
    def __init__(
        self, input_info: ModelInputInfo, fwd_signature: Signature, module_ref_for_device: torch.nn.Module = None
    ):
        self._fwd_signature = fwd_signature
        self._module_ref_for_device = module_ref_for_device
        args, kwargs = input_info.get_forward_inputs()
        bound_params = fwd_signature.bind(*args, **kwargs)
        self._fwd_param_names_to_dummy_inputs_odict: Dict[str, Any] = bound_params.arguments

    def wrap_inputs(self, model_args: Tuple, model_kwargs: Dict) -> Tuple[Tuple, Dict]:
        bound_model_params = self._fwd_signature.bind(*model_args, **model_kwargs)
        for param_name, dummy_input in self._fwd_param_names_to_dummy_inputs_odict.items():
            if not isinstance(dummy_input, torch.Tensor):
                continue  # Did not expect a tensor at this parameter during graph building, shouldn't wrap now too

            param_kind = self._fwd_signature.parameters[param_name].kind
            if param_kind is Parameter.VAR_POSITIONAL or param_kind is Parameter.VAR_KEYWORD:
                nncf_logger.warning(
                    "An input_info tensor was bound to a *args or **kwargs variadic parameter in the"
                    "forward's signature! This is currently unsupported by NNCF. Input compression may "
                    "be incorrect."
                )
                # Currently won't support input info mapping to *args or **kwargs-mapped parameters
                continue

            if param_name not in bound_model_params.arguments:
                nncf_logger.warning(
                    "A call to a compressed model's forward occurred without one of the arguments "
                    "specified at the compressed model creation stage! Input compression may be incorrect. "
                    "Trying to recover by wrapping the default value for the argument."
                )
                bound_model_params.apply_defaults()

            potential_tensor = bound_model_params.arguments[param_name]

            if potential_tensor is None:
                # Default was None - cannot wrap as-is. Will wrap a dummy tensor as specified in
                # input info - will preserve the call order of nncf_model_input nodes,
                # and the post-hooks for the input node will execute. The result won't go anywhere, though.
                nncf_logger.info(f"Wrapping a dummy tensor for input {param_name}")
                device = "cuda"
                if self._module_ref_for_device is not None:
                    device = get_model_device(self._module_ref_for_device)
                dummy_input_copy = dummy_input.clone().to(device)
                _ = nncf_model_input(dummy_input_copy)
            elif isinstance(potential_tensor, torch.Tensor):
                # Skip wrapping by nncf_model_input in case potential tensor is not a torch.Tensor.
                bound_model_params.arguments[param_name] = nncf_model_input(potential_tensor)

        return bound_model_params.args, bound_model_params.kwargs


class HasDataloader(Protocol):
    @property
    def data_loader(self) -> NNCFDataLoader:
        pass


EXTRA_STRUCTS_WITH_DATALOADERS: List[Type[HasDataloader]] = [QuantizationRangeInitArgs, BNAdaptationInitArgs]



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import Callable, List, Optional, TypeVar

from nncf import Dataset
from nncf.common.graph.graph import NNCFGraph
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.experimental.quantization.algorithms.post_training.pipeline import experimental_create_ptq_pipeline
from nncf.experimental.quantization.quantizer import Quantizer
from nncf.quantization.advanced_parameters import AdvancedBiasCorrectionParameters
from nncf.quantization.advanced_parameters import AdvancedSmoothQuantParameters
from nncf.quantization.advanced_parameters import RangeEstimatorParameters
from nncf.quantization.algorithms.algorithm import Algorithm

TModel = TypeVar("TModel")
TPass = Callable[[TModel], TModel]


class ExperimentalPostTrainingQuantization(Algorithm):
    """
    Implements Experimental Post-Training Quantization algorithm, which basically includes:
    1) ChannelAlignment
    2) MinMaxRangeInit
    3) FastBiasCorrection or BiasCorrection
    """

    def __init__(
        self,
        quantizer: Quantizer,
        subset_size: int = 300,
        fast_bias_correction: Optional[bool] = True,
        smooth_quant: bool = False,
        bias_correction_params: Optional[AdvancedBiasCorrectionParameters] = None,
        smooth_quant_params: Optional[AdvancedSmoothQuantParameters] = None,
        activations_range_estimator_params: Optional[RangeEstimatorParameters] = None,
        weights_range_estimator_params: Optional[RangeEstimatorParameters] = None,
        batchwise_statistics: bool = False,
    ):
        """
        :param quantizer: Quantizer to use in MiMaxRangeInit algorithm.
        :param subset_size: Size of a subset to calculate activations
            statistics used for quantization.
        :param fast_bias_correction: Setting this option to `False` enables a different
            bias correction method which is more accurate, in general, and takes
            more time but requires less memory. None disables the bias correction algorithm.
        :param smooth_quant: Setting this option to `True` enables the SmoothQuant algorithm.
        :param bias_correction_params: Contains advanced parameters for fine-tuning bias correction algorithm.
        :param smooth_quant_params: Contains advanced alpha parameters for SmoothQuant algorithm.
        :param activations_range_estimator_params: Contains parameters for estimating the range
            of activations of the model.
        :param weights_range_estimator_params: Contains parameters for estimating the range
            of weights of the model.
        :param batchwise_statistics: Determines whether quantizer statistics should be calculated
            for each item of the batch or for the entire batch, default is False.
        """
        self._pipeline = experimental_create_ptq_pipeline(
            quantizer=quantizer,
            subset_size=subset_size,
            fast_bias_correction=fast_bias_correction,
            smooth_quant=smooth_quant,
            bias_correction_params=bias_correction_params,
            smooth_quant_params=smooth_quant_params,
            activations_range_estimator_params=activations_range_estimator_params,
            weights_range_estimator_params=weights_range_estimator_params,
            batchwise_statistics=batchwise_statistics,
        )

    @property
    def available_backends(self) -> List[BackendType]:
        backends = set(BackendType)
        for algorithm in itertools.chain.from_iterable(self._pipeline.pipeline_steps):
            backends = backends.intersection(algorithm.available_backends)
        return list(backends)

    def get_statistic_points(self, model: TModel, graph: NNCFGraph) -> StatisticPointsContainer:
        return self._pipeline.get_statistic_points_for_step(0, model, graph)

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> TModel:
        if dataset is None and len(self._pipeline.pipeline_steps) > 1:
            raise ValueError(
                "A dataset is required for the post-training quantization "
                "algorithm to collect statistics for intermediate models."
            )

        step_index_to_statistics = None
        if statistic_points:
            step_index_to_statistics = {0: statistic_points}

        return self._pipeline.run_from_step(model, dataset, graph, 0, step_index_to_statistics)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import functools
import inspect
from contextlib import contextmanager
from typing import Callable, List, Optional, Union

import torch
import torch.utils.cpp_extension
from torch._jit_internal import createResolutionCallbackFromFrame
from torch.jit import is_tracing
from torch.nn import DataParallel
from torch.nn.parallel import DistributedDataParallel

import nncf
from nncf import nncf_logger
from nncf.common.utils.api_marker import api
from nncf.torch.dynamic_graph.patch_pytorch_state import PATCHING_STATE
from nncf.torch.dynamic_graph.structs import NamespaceTarget
from nncf.torch.dynamic_graph.structs import PatchedOperatorInfo
from nncf.torch.dynamic_graph.trace_tensor import TracedParameter
from nncf.torch.dynamic_graph.trace_tensor import TracedTensor
from nncf.torch.dynamic_graph.wrappers import ignore_scope
from nncf.torch.dynamic_graph.wrappers import wrap_module_call
from nncf.torch.dynamic_graph.wrappers import wrap_operator


def get_namespaces_to_patch(namespace_target: NamespaceTarget) -> object:
    if namespace_target == NamespaceTarget.TORCH_NN_FUNCTIONAL:
        return torch.nn.functional
    if namespace_target == NamespaceTarget.TORCH_TENSOR:
        return TracedTensor
    if namespace_target == NamespaceTarget.TORCH_NN_PARAMETER:
        return TracedParameter
    if namespace_target == NamespaceTarget.TORCH:
        return torch
    raise nncf.ValidationError("{} namespace wasn't found in {}".format(namespace_target, NamespaceTarget))


def get_namespace_to_extract_functions_from(namespace_target: NamespaceTarget) -> object:
    if namespace_target == NamespaceTarget.TORCH_NN_FUNCTIONAL:
        return torch.nn.functional
    if namespace_target == NamespaceTarget.TORCH_TENSOR:
        return torch.Tensor
    if namespace_target == NamespaceTarget.TORCH_NN_PARAMETER:
        return torch.nn.Parameter
    if namespace_target == NamespaceTarget.TORCH:
        return torch._C._VariableFunctions
    raise nncf.ValidationError("{} namespace wasn't found in {}".format(namespace_target, NamespaceTarget))


class FunctionsToPatchWithoutTracing:
    TENSOR_CREATING_FUNCTIONS = [
        "arange",
        "as_subclass",
        "as_tensor",
        "copysign",
        "copysign_",
        "empty",
        "ones",
        "ones_like",
        "rad2deg",
        "rad2deg_",
        "rand",
        "randn",
        "randn_like",
        "tensor",
        "zeros",
    ]
    TENSOR_UTILITY_FUNCTIONS = [
        "all",
        "allclose",
        "any",
        "assert_int_or_pair",
        "backward",
        "broadcast_to",
        "cpu",
        "cuda",
        "data_ptr",
        "dequantize",
        "dim",
        "handle_torch_function",
        "has_names",
        "has_torch_function",
        "has_torch_function_unary",
        "has_torch_function_variadic",
        "is_contiguous",
        "item",
        "names",
        "numel",
        "numpy",
        "q_per_channel_axis",
        "q_per_channel_scales",
        "q_per_channel_zero_points",
        "q_scale",
        "q_zero_point",
        "qr",
        "qscheme",
        "random_",
        "record_stream",
        "refine_names",
        "register_hook",
        "rename",
        "rename_",
        "shape",
        "size",
        "sort",
        "storage",
        "storage_offset",
        "stride",
        "get_device",
        "is_floating_point",
    ]

    FUNCTIONS_TO_PATCH_WITHOUT_TRACING = TENSOR_CREATING_FUNCTIONS + TENSOR_UTILITY_FUNCTIONS


class MagicFunctionsToPatch:
    TENSOR_MAGIC_FUNCTIONS = [
        "__abs__",
        "__add__",
        "__and__",
        "__div__",
        "__eq__",
        "__floordiv__",
        "__ge__",
        "__getitem__",
        "__gt__",
        "__iadd__",
        "__iand__",
        "__idiv__",
        "__ifloordiv__",
        "__imul__",
        "__invert__",
        "__ior__",
        "__ipow__",
        "__isub__",
        "__itruediv__",
        "__ixor__",
        "__le__",
        "__lt__",
        "__matmul__",
        "__mod__",
        "__mul__",
        "__ne__",
        "__neg__",
        "__or__",
        "__pow__",
        "__radd__",
        "__rand__",
        "__rdiv__",
        "__rfloordiv__",
        "__rmatmul__",
        "__rmul__",
        "__ror__",
        "__rpow__",
        "__rsub__",
        "__rtruediv__",
        "__rxor__",
        "__sub__",
        "__truediv__",
        "__xor__",
    ]
    MAGIC_FUNCTIONS_TO_PATCH = {
        NamespaceTarget.TORCH_TENSOR: TENSOR_MAGIC_FUNCTIONS,
        NamespaceTarget.TORCH_NN_PARAMETER: TENSOR_MAGIC_FUNCTIONS + ["get_dtype"],
    }


@api(canonical_alias="nncf.torch.register_operator")
def register_operator(name=None):
    def wrap(operator):
        op_name = name
        if op_name is None:
            op_name = operator.__name__
        return wrap_operator(operator, PatchedOperatorInfo(op_name, NamespaceTarget.EXTERNAL))

    return wrap

    # TODO: Use same wrapper for model.forward() calls


def torch_jit_script_wrapper(*args, **kwargs):
    # Torch JIT cannot work with NNCF-modified operators,
    # so at call of torch.jit.script function we need to
    # un-patch the torch operators

    with disable_patching():
        signature = inspect.signature(_ORIG_JIT_SCRIPT)
        bound_args = signature.bind(*args, **kwargs).arguments
        # Process the case when the object-to-script is a class as in the original jit.script logic
        if inspect.isclass(bound_args["obj"]):
            # Inserting wrapper alters the call stack, hence we need to change the resolution callback accordingly
            if "_rcb" not in bound_args:
                frames_up = bound_args.get("_frames_up", 0)
                rcb = createResolutionCallbackFromFrame(frames_up + 1)
                kwargs["_rcb"] = rcb
            retval = _ORIG_JIT_SCRIPT(*args, **kwargs)
        else:
            # For some reason resolution callback may return patched methods, so we wrap it to avoid this
            if "_rcb" in kwargs:
                rcb = kwargs["_rcb"]

                def rcb_wrapper(name):
                    value = rcb(name)
                    if hasattr(value, "_original_op"):
                        value = value._original_op
                    return value

                kwargs["_rcb"] = rcb_wrapper

            retval = _ORIG_JIT_SCRIPT(*args, **kwargs)

        return retval


def torch_jit_trace_make_module_wrapper(*args, **kwargs):
    with disable_patching():
        retval = _ORIG_JIT_TRACE_MAKE_MODULE(*args, **kwargs)
    return retval


def torch_jit_script_if_tracing(fn):
    @functools.wraps(fn)
    def wrapper(*args, **kwargs):
        if not is_tracing():
            return fn(*args, **kwargs)

        compiled_fn = torch.jit.script(wrapper.__original_fn)
        return compiled_fn(*args, **kwargs)

    wrapper.__original_fn = fn
    wrapper.__script_if_tracing_wrapper = True

    return wrapper


def get_torch_compile_wrapper():
    """
    Wrapper for torch.compile() that disables NNCF patching when called for vanilla PyTorch model and
    raises an exception when called for an NNCF-optimized model.
    """

    @functools.wraps(_ORIG_TORCH_COMPILE)
    def wrapper(model: Optional[Callable] = None, **kwargs):
        from nncf.torch.nncf_network import NNCFNetwork

        if isinstance(model, NNCFNetwork):
            raise TypeError("At the moment torch.compile() is not supported for models optimized by NNCF.")
        with disable_patching():
            return _ORIG_TORCH_COMPILE(model, **kwargs)

    return wrapper


class OriginalOpInfo:
    def __init__(self, name: str, namespace, op):
        self.name = name
        self.namespace = namespace
        self.op = op


ORIGINAL_OPERATORS: List[OriginalOpInfo] = []
ORIGINAL_CALL = torch.nn.Module.__call__
_ORIG_JIT_SCRIPT = None
_ORIG_JIT_TRACE_MAKE_MODULE = None
_ORIG_TORCH_COMPILE: Union[Callable, None] = None


@functools.wraps(ORIGINAL_CALL)
def unpatching_module_call(*args, **kwargs):
    # Wrapper for module.__call__ that unpatches torch operators during model forward
    with disable_patching():
        return ORIGINAL_CALL(*args, **kwargs)


def patch_torch_jit():
    # This import statement is required, otherwise we get a
    # "RuntimeError: undefined value torch" inside the real torch.jit.script

    import torch

    global _ORIG_JIT_SCRIPT
    _ORIG_JIT_SCRIPT = getattr(torch.jit, "script")
    setattr(torch.jit, "script", torch_jit_script_wrapper)

    # Patch torch.jit._trace.make_module() which is called during
    # torch.jit.trace() call
    global _ORIG_JIT_TRACE_MAKE_MODULE
    _ORIG_JIT_TRACE_MAKE_MODULE = getattr(torch.jit._trace, "make_module")
    setattr(torch.jit._trace, "make_module", torch_jit_trace_make_module_wrapper)

    # Patch torch.jit._script_if_tracing because it references an original
    # unpatched torch.jit.script and the patching above does not affect it
    setattr(torch.jit, "_script_if_tracing", torch_jit_script_if_tracing)


def patch_namespace_opname(namespace, op_info: PatchedOperatorInfo):
    op_name = op_info.name
    if hasattr(namespace, op_name):
        orig = getattr(namespace, op_name)
        ORIGINAL_OPERATORS.append(OriginalOpInfo(op_name, namespace, orig))
        setattr(namespace, op_name, wrap_operator(orig, op_info))
    else:
        nncf_logger.debug(f"Not patching {op_name} since it is missing in this version of PyTorch")


def get_all_functions_from_namespace(namespace: NamespaceTarget, do_filter: bool = True) -> List[str]:
    """
    Seeks all attributes from the namespace, then takes only attributes,
    which types are function, builtin, method or method descriptor.
    If 'do_filer' is True, then also removes all private or magic attributes.
    :param namespace: Python module.
    :param do_filter: If True return only public functions, else - otherwise.
    """

    def remove_private_functions(names: List[str]) -> List[str]:
        filtered_names = []
        for name in names:
            if name.startswith("_"):
                continue
            filtered_names.append(name)
        return filtered_names

    patched_namespace = get_namespace_to_extract_functions_from(namespace)
    all_torch_function_names = []
    members = inspect.getmembers(patched_namespace)
    for member in members:
        if (
            inspect.isfunction(member[1])
            or inspect.isbuiltin(member[1])
            or inspect.ismethod(member[1])
            or inspect.ismethoddescriptor(member[1])
        ):
            all_torch_function_names.append(member[0])
    if do_filter:
        filtered_function_names = remove_private_functions(all_torch_function_names)
        return filtered_function_names
    return all_torch_function_names


def patch_torch_operators():
    # Only patch torch.jit.script during first patch_torch_operators call
    if not PATCHING_STATE.jit_is_wrapped:
        patch_torch_jit()
        PATCHING_STATE.jit_is_wrapped = True

    # Unpatch torch operators during model compilation.
    if not PATCHING_STATE.compile_is_wrapped:
        global _ORIG_TORCH_COMPILE
        _ORIG_TORCH_COMPILE = torch.compile
        setattr(torch, "compile", get_torch_compile_wrapper())
        PATCHING_STATE.compile_is_wrapped = True

    # Do not patch operators twice as well
    if PATCHING_STATE.operators_are_wrapped:
        return
    PATCHING_STATE.operators_are_wrapped = True

    global ORIGINAL_OPERATORS
    ORIGINAL_OPERATORS = []

    functions_to_patch = {}
    for namespace in NamespaceTarget:
        if namespace in [NamespaceTarget.ATEN, NamespaceTarget.EXTERNAL]:
            continue
        functions_to_patch[namespace] = get_all_functions_from_namespace(namespace)

    ignored_functions = FunctionsToPatchWithoutTracing.FUNCTIONS_TO_PATCH_WITHOUT_TRACING
    functions_to_patch_without_tracing = {}
    for namespace, function_names in functions_to_patch.items():
        functions_to_patch_without_tracing[namespace] = []
        new_function_names = []
        for function_name in function_names:
            if function_name in ignored_functions:
                functions_to_patch_without_tracing[namespace].append(function_name)
                continue
            new_function_names.append(function_name)
        functions_to_patch[namespace] = new_function_names

    # Relu function from torch.nn.functional uses torch.relu or torch.relu_ inside,
    # which we've already wrapped. So we don't have to wrap Relu in torch.nn.functional
    # to be able to differ what torch.relu or torch.relu was called exactly inside Relu
    functions_to_patch[NamespaceTarget.TORCH_NN_FUNCTIONAL].remove("relu")

    # nn.MultiheadAttention uses multi_head_attention_forward from torch.nn.functional inside, which
    # leads to adding a single node to the graph. In turn, it makes impossible to quantize internals of the multi head
    # attention which is required for better speed-up. So multi_head_attention_forward is skipped from wrapping and
    # tracing goes inside and finds internal operations in it: bmm, linear, softmax and etc.
    functions_to_patch[NamespaceTarget.TORCH_NN_FUNCTIONAL].remove("multi_head_attention_forward")

    magic_functions_to_patch = MagicFunctionsToPatch.MAGIC_FUNCTIONS_TO_PATCH
    for namespace, function_names in magic_functions_to_patch.items():
        functions_to_patch[namespace] += function_names

    for namespace, function_names in functions_to_patch.items():
        for function_name in function_names:
            op_info = PatchedOperatorInfo(function_name, namespace)
            patched_namespace = get_namespaces_to_patch(namespace)
            patch_namespace_opname(patched_namespace, op_info)

    # Patch operators without tracing so that
    # both they and any internal calls to otherwise traced functions do not appear into the model graph.

    for namespace, function_names in functions_to_patch_without_tracing.items():
        for function_name in function_names:
            op_info = PatchedOperatorInfo(function_name, namespace, skip_trace=True)
            patched_namespace = get_namespaces_to_patch(namespace)
            patch_namespace_opname(patched_namespace, op_info)

    # Patch __repr__ twice in 'torch.Tensor' and 'TracedTensor'.
    # This is done to not add operations behind print() operator for the both TracedTensor and torch.Tensor.

    op_info = PatchedOperatorInfo("__repr__", NamespaceTarget.TORCH_TENSOR, skip_trace=True)
    patch_namespace_opname(torch.Tensor, op_info)
    op_info = PatchedOperatorInfo("__repr__", NamespaceTarget.TORCH_TENSOR, skip_trace=True)
    patch_namespace_opname(TracedTensor, op_info)

    ORIGINAL_OPERATORS.append(OriginalOpInfo("__call__", torch.nn.Module, torch.nn.Module.__call__))
    torch.nn.Module.__call__ = wrap_module_call(torch.nn.Module.__call__)
    ignore_scope(DataParallel)
    ignore_scope(DistributedDataParallel)


def unpatch_torch_operators():
    if not PATCHING_STATE.operators_are_wrapped:
        return
    PATCHING_STATE.operators_are_wrapped = False

    for orig_op_info in ORIGINAL_OPERATORS:
        setattr(orig_op_info.namespace, orig_op_info.name, orig_op_info.op)


@contextmanager
def disable_patching():
    was_patched = PATCHING_STATE.operators_are_wrapped
    if was_patched:
        unpatch_torch_operators()
    try:
        yield
    finally:
        # The code in the with statement may raise an exception, which could be expected to be handled elsewhere.
        # Need to restore the previous state of patching in this case before continuing to the exception handling.
        if was_patched:
            patch_torch_operators()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from enum import Enum
from typing import Any, Optional, Union

SerializableData = Union[str, Enum, bool]


@dataclass
class CollectedEvent:
    """
    name: The name for the event, will be sent as `event_action`
    data: Optional - the data associated with the event. Must be a string - serialize to string if it is not.
    int_data: Optional - the integer data associated with the event. Must be a positive integer.
    """

    name: str
    data: Optional[SerializableData] = None  # GA limitations
    int_data: Optional[int] = None


class TelemetryExtractor(ABC):
    """
    Interface for custom telemetry extractors, to be used with the `nncf.telemetry.tracked_function` decorator.
    """

    def __init__(self, argname: str = ""):
        self._argname = argname

    @property
    def argname(self) -> Optional[str]:
        return self._argname

    @abstractmethod
    def extract(self, argvalue: Any) -> CollectedEvent:
        """
        Implement this method to prepare the telemetry event data from the tracked function's argument value
        passed via `argvalue`.
        """


class VerbatimTelemetryExtractor(TelemetryExtractor):
    def extract(self, argvalue: SerializableData) -> CollectedEvent:
        if isinstance(argvalue, Enum):
            argvalue = str(argvalue.value)
        if isinstance(argvalue, bool):
            argvalue = "enabled" if argvalue else "disabled"
        return CollectedEvent(name=self._argname, data=argvalue)


class FunctionCallTelemetryExtractor(TelemetryExtractor):
    def __init__(self, argvalue: Any = None) -> None:
        super().__init__()
        self._argvalue = argvalue

    def extract(self, _: Any) -> CollectedEvent:
        return CollectedEvent(name="function_call", data=self._argvalue)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Dynamic Version Generation for Package Builds

This module is responsible for generating the dynamic version for the package during the build process.
It provides the `version` attribute for setuptools as specified in the `pyproject.toml` configuration file:

    [tool.setuptools.dynamic]
    version = { attr = "_version_helper.version" }

Environment Variable Instructions:
-----------------------------------
1. **NNCF_RELEASE_BUILD**:
    - Set this environment variable to generate a release package using `python -m build` without including
      the commit hash in the version.
    - If this variable is not set, the file `nncf/version.py` will be overridden with a custom version
      that includes the commit hash. Example usage:

        NNCF_RELEASE_BUILD=1 python -m build

2. **NNCF_BUILD_SUFFIX**:
    - Use this environment variable to specify a particular suffix for the build. Example usage:

        NNCF_BUILD_SUFFIX="rc1" python -m build

Post-Build Recommendation:
---------------------------
After generating the package, it is recommended to revert any changes to `nncf/version.py` to avoid potential conflicts.
This can be done using the following command:

    git checkout nncf/version.py

This ensures that `nncf/version.py` remains in its original state after the dynamic versioning process.
"""


from __future__ import annotations

import contextlib
import os
import re
import subprocess
from pathlib import Path

NNCF_VERSION_FILE = "nncf/version.py"


def get_custom_version() -> str:
    version_match = re.search(
        r"^__version__ = ['\"]((\d+\.\d+\.\d+)([^'\"]*))['\"]", Path(NNCF_VERSION_FILE).read_text(), re.M
    )
    if not version_match:
        raise RuntimeError("Unable to find version string.")

    version_full = version_match.group(1)
    version_value = version_match.group(2)
    version_suffix = version_match.group(3)

    nncf_build_suffix = os.environ.get("NNCF_BUILD_SUFFIX")
    if nncf_build_suffix:
        # Suffix expected on build package
        return f"{version_value}.{nncf_build_suffix}"

    is_building_release = "NNCF_RELEASE_BUILD" in os.environ
    if is_building_release or version_suffix:
        return version_full

    dev_version_id = "unknown_version"
    repo_root = os.path.dirname(os.path.realpath(__file__))

    # Get commit hash
    with contextlib.suppress(subprocess.CalledProcessError):
        dev_version_id = (
            subprocess.check_output(["git", "rev-parse", "--short", "HEAD"], cwd=repo_root).strip().decode()  # nosec
        )

    # Detect modified files
    with contextlib.suppress(subprocess.CalledProcessError):
        run = subprocess.run(["git", "diff-index", "--quiet", "HEAD"], cwd=repo_root)  # nosec
        if run.returncode == 1:
            dev_version_id += "dirty"

    return version_value + f".dev0+{dev_version_id}"


version: str


def __getattr__(name: str) -> str:

    if name == "version":
        global version
        version = get_custom_version()

        # Rewrite version.py to pass custom version to package
        if os.environ.get("_PYPROJECT_HOOKS_BUILD_BACKEND"):
            content = Path(NNCF_VERSION_FILE).read_text()
            version_str = re.search(r"^__version__ = ['\"][^'\"]*['\"]", content, re.M).group(0)
            content = content.replace(version_str, f'__version__ = "{version}"')
            Path(NNCF_VERSION_FILE).write_text(content)

        return version

    raise AttributeError(name)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Type

from nncf.common.graph.operator_metatypes import OperatorMetatype
from nncf.common.hardware.config import HWConfig
from nncf.torch.graph.operator_metatypes import get_operator_metatypes


class PTHWConfig(HWConfig):
    def _get_available_operator_metatypes_for_matching(self) -> List[Type[OperatorMetatype]]:
        return get_operator_metatypes()



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.torch.compression_method_api import PTCompressionAlgorithmController


class QuantizationControllerBase(PTCompressionAlgorithmController):
    """
    Base controller class for the quantization controllers in PT.
    """

    def enable_activation_quantization(self):
        raise NotImplementedError

    def enable_weight_quantization(self):
        raise NotImplementedError

    def disable_activation_quantization(self):
        raise NotImplementedError

    def disable_weight_quantization(self):
        raise NotImplementedError

    def init_range(self):
        raise NotImplementedError



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.common.graph.patterns.patterns import GraphPattern
from nncf.common.graph.patterns.patterns import IgnoredPatternNames
from nncf.common.utils.registry import Registry
from nncf.torch.graph import operator_metatypes as om
from nncf.torch.graph.pattern_operations import ATOMIC_ACTIVATIONS_OPERATIONS
from nncf.torch.graph.pattern_operations import LINEAR_OPERATIONS

PT_IGNORED_PATTERNS = Registry("IGNORED_PATTERNS")


def _add_softmax_matmul(
    pattern: GraphPattern,
    matmul_metatypes,
    reshape_squeeze_metatypes,
    gather_metatypes,
    transpose_metatypes,
    concat_metatypes,
) -> None:
    #       SOFTMAX  RESHAPE||TRANSPOSE||GATHER||SQUEEZE||CONCAT
    #           \              /
    #            \            /
    #             \          /
    #              \        /
    #               \      /
    #                MATMUL
    branch_matmul_nodes = reshape_squeeze_metatypes + gather_metatypes + transpose_metatypes + concat_metatypes
    softmax = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SOFTMAX", GraphPattern.METATYPE_ATTR: om.PTSoftmaxMetatype})
    matmul = pattern.add_node(**{GraphPattern.LABEL_ATTR: "MATMUL", GraphPattern.METATYPE_ATTR: matmul_metatypes})
    matmul_branch_nodes = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "NON_PATTERN", GraphPattern.METATYPE_ATTR: branch_matmul_nodes}
    )
    pattern.add_edge(softmax, matmul)
    pattern.add_edge(matmul_branch_nodes, matmul)


def _add_softmax_reshape_matmul(
    pattern: GraphPattern,
    matmul_metatypes,
    reshape_squeeze_metatypes,
    gather_metatypes,
    transpose_metatypes,
    concat_metatypes,
) -> None:
    #       SOFTMAX
    #           \
    #            \
    #             \
    #    RESHAPE || TRANSPOSE   RESHAPE||TRANSPOSE||GATHER||SQUEEZE||CONCAT
    #                 \                 /
    #                  \               /
    #                   \             /
    #                    \           /
    #                     \         /
    #                      \       /
    #                        MATMUL
    branch_matmul_nodes = reshape_squeeze_metatypes + gather_metatypes + transpose_metatypes + concat_metatypes
    softmax = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SOFTMAX", GraphPattern.METATYPE_ATTR: om.PTSoftmaxMetatype})
    reshape = pattern.add_node(
        **{
            GraphPattern.LABEL_ATTR: "RESHAPE",
            GraphPattern.METATYPE_ATTR: reshape_squeeze_metatypes + transpose_metatypes,
        }
    )
    matmul = pattern.add_node(**{GraphPattern.LABEL_ATTR: "MATMUL", GraphPattern.METATYPE_ATTR: matmul_metatypes})
    matmul_branch_nodes = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "RESHAPE||TRANSPOSE||GATHER", GraphPattern.METATYPE_ATTR: branch_matmul_nodes}
    )
    pattern.add_edge(softmax, reshape)
    pattern.add_edge(reshape, matmul)
    pattern.add_edge(matmul_branch_nodes, matmul)
    return pattern


@PT_IGNORED_PATTERNS.register(IgnoredPatternNames.MULTIHEAD_ATTENTION_OUTPUT)
def create_multihead_attention_output() -> GraphPattern:
    matmul_metatypes = [om.PTLinearMetatype, om.PTAddmmMetatype, om.PTMatMulMetatype]
    reshape_squeeze_metatypes = [om.PTReshapeMetatype, om.PTSqueezeMetatype, om.PTSplitMetatype]
    gather_metatypes = [om.PTGatherMetatype]
    transpose_metatypes = [om.PTTransposeMetatype]
    concat_metatypes = [om.PTCatMetatype]

    pattern = GraphPattern()
    _add_softmax_matmul(
        pattern,
        matmul_metatypes=matmul_metatypes,
        reshape_squeeze_metatypes=reshape_squeeze_metatypes,
        gather_metatypes=gather_metatypes,
        transpose_metatypes=transpose_metatypes,
        concat_metatypes=concat_metatypes,
    )
    _add_softmax_reshape_matmul(
        pattern,
        matmul_metatypes=matmul_metatypes,
        reshape_squeeze_metatypes=reshape_squeeze_metatypes,
        gather_metatypes=gather_metatypes,
        transpose_metatypes=transpose_metatypes,
        concat_metatypes=concat_metatypes,
    )
    return pattern


# pylint:disable=too-many-statements
@PT_IGNORED_PATTERNS.register(IgnoredPatternNames.SE_BLOCK)
def create_se_block() -> GraphPattern:
    MEAN_OPERATIONS = {
        GraphPattern.LABEL_ATTR: "REDUCE_MEAN",
        GraphPattern.METATYPE_ATTR: [om.PTAvgPool2dMetatype, om.PTAvgPool3dMetatype, om.PTMeanMetatype],
        GraphPattern.PATTERN_NODE_TO_EXCLUDE: True,
    }
    SYGMOID_OPERATIONS = {
        GraphPattern.LABEL_ATTR: "SIGMOID",
        GraphPattern.METATYPE_ATTR: [om.PTSigmoidMetatype, om.PTHardSigmoidMetatype],
    }
    MUL_OPERATION = {
        GraphPattern.LABEL_ATTR: "MUL",
        GraphPattern.METATYPE_ATTR: om.PTMulMetatype,
        GraphPattern.PATTERN_NODE_TO_EXCLUDE: True,
    }

    def get_se_block_pattern() -> GraphPattern:
        pattern = GraphPattern()
        any_node = pattern.add_node(label="NON_PATTERN_NODE", type=GraphPattern.NON_PATTERN_NODE_TYPE)
        reduce_mean_node = pattern.add_node(**MEAN_OPERATIONS)
        linear_node_1 = pattern.add_node(**LINEAR_OPERATIONS)
        activation_node_1 = pattern.add_node(**ATOMIC_ACTIVATIONS_OPERATIONS)
        linear_node_2 = pattern.add_node(**LINEAR_OPERATIONS)
        activation_node_2 = pattern.add_node(**SYGMOID_OPERATIONS)
        multiply_node = pattern.add_node(**MUL_OPERATION)

        pattern.add_edge(any_node, reduce_mean_node)
        pattern.add_edge(reduce_mean_node, linear_node_1)
        pattern.add_edge(linear_node_1, activation_node_1)
        pattern.add_edge(activation_node_1, linear_node_2)
        pattern.add_edge(linear_node_2, activation_node_2)
        pattern.add_edge(activation_node_2, multiply_node)
        pattern.add_edge(any_node, multiply_node)
        return pattern

    def get_se_block_with_bias_pattern() -> GraphPattern:
        pattern = GraphPattern()
        any_node = pattern.add_node(label="NON_PATTERN_NODE", type=GraphPattern.NON_PATTERN_NODE_TYPE)
        reduce_mean_node = pattern.add_node(**MEAN_OPERATIONS)
        linear_node_1 = pattern.add_node(**LINEAR_OPERATIONS)
        add_node_1 = pattern.add_node(label="ADD_BIAS", type=[om.PTAddMetatype, om.PTSubMetatype])
        activation_node_1 = pattern.add_node(**ATOMIC_ACTIVATIONS_OPERATIONS)
        linear_node_2 = pattern.add_node(**LINEAR_OPERATIONS)
        add_node_2 = pattern.add_node(label="ADD_BIAS", type=[om.PTAddMetatype, om.PTSubMetatype])
        activation_node_2 = pattern.add_node(**SYGMOID_OPERATIONS)
        multiply_node = pattern.add_node(**MUL_OPERATION)

        pattern.add_edge(any_node, reduce_mean_node)
        pattern.add_edge(reduce_mean_node, linear_node_1)
        pattern.add_edge(linear_node_1, add_node_1)
        pattern.add_edge(add_node_1, activation_node_1)
        pattern.add_edge(activation_node_1, linear_node_2)
        pattern.add_edge(linear_node_2, add_node_2)
        pattern.add_edge(add_node_2, activation_node_2)
        pattern.add_edge(activation_node_2, multiply_node)
        pattern.add_edge(any_node, multiply_node)
        return pattern

    RESHAPE_NODES = {
        GraphPattern.LABEL_ATTR: "RESHAPE",
        GraphPattern.METATYPE_ATTR: om.PTReshapeMetatype,
    }

    def get_se_block_with_reshape() -> GraphPattern:
        pattern = GraphPattern()
        any_node = pattern.add_node(label="NON_PATTERN_NODE", type=GraphPattern.NON_PATTERN_NODE_TYPE)
        reduce_mean_node = pattern.add_node(**MEAN_OPERATIONS)
        reshape_node_1 = pattern.add_node(**RESHAPE_NODES)
        linear_node_1 = pattern.add_node(**LINEAR_OPERATIONS)
        activation_node_1 = pattern.add_node(**ATOMIC_ACTIVATIONS_OPERATIONS)
        linear_node_2 = pattern.add_node(**LINEAR_OPERATIONS)
        activation_node_2 = pattern.add_node(**SYGMOID_OPERATIONS)
        reshape_node_2 = pattern.add_node(**RESHAPE_NODES)
        multiply_node = pattern.add_node(**MUL_OPERATION)

        pattern.add_edge(any_node, reduce_mean_node)
        pattern.add_edge(reduce_mean_node, reshape_node_1)
        pattern.add_edge(reshape_node_1, linear_node_1)
        pattern.add_edge(linear_node_1, activation_node_1)
        pattern.add_edge(activation_node_1, linear_node_2)
        pattern.add_edge(linear_node_2, activation_node_2)
        pattern.add_edge(activation_node_2, reshape_node_2)
        pattern.add_edge(reshape_node_2, multiply_node)
        pattern.add_edge(any_node, multiply_node)
        return pattern

    def get_se_block_with_bias_and_reshape() -> GraphPattern:
        pattern = GraphPattern()
        any_node = pattern.add_node(label="NON_PATTERN_NODE", type=GraphPattern.NON_PATTERN_NODE_TYPE)
        reduce_mean_node = pattern.add_node(**MEAN_OPERATIONS)
        reshape_node_1 = pattern.add_node(**RESHAPE_NODES)
        linear_node_1 = pattern.add_node(**LINEAR_OPERATIONS)
        add_node_1 = pattern.add_node(label="ADD_BIAS", type=[om.PTAddMetatype, om.PTSubMetatype])
        activation_node_1 = pattern.add_node(**ATOMIC_ACTIVATIONS_OPERATIONS)
        linear_node_2 = pattern.add_node(**LINEAR_OPERATIONS)
        add_node_2 = pattern.add_node(label="ADD_BIAS", type=[om.PTAddMetatype, om.PTSubMetatype])
        activation_node_2 = pattern.add_node(**SYGMOID_OPERATIONS)
        reshape_node_2 = pattern.add_node(**RESHAPE_NODES)
        multiply_node = pattern.add_node(**MUL_OPERATION)

        pattern.add_edge(any_node, reduce_mean_node)
        pattern.add_edge(reduce_mean_node, reshape_node_1)
        pattern.add_edge(reshape_node_1, linear_node_1)
        pattern.add_edge(linear_node_1, add_node_1)
        pattern.add_edge(add_node_1, activation_node_1)
        pattern.add_edge(activation_node_1, linear_node_2)
        pattern.add_edge(linear_node_2, add_node_2)
        pattern.add_edge(add_node_2, activation_node_2)
        pattern.add_edge(activation_node_2, reshape_node_2)
        pattern.add_edge(reshape_node_2, multiply_node)
        pattern.add_edge(any_node, multiply_node)
        return pattern

    main_pattern = GraphPattern()
    main_pattern.add_pattern_alternative(get_se_block_pattern())
    main_pattern.add_pattern_alternative(get_se_block_with_bias_pattern())
    main_pattern.add_pattern_alternative(get_se_block_with_reshape())
    main_pattern.add_pattern_alternative(get_se_block_with_bias_and_reshape())
    return main_pattern


@PT_IGNORED_PATTERNS.register(IgnoredPatternNames.ROPE)
def create_rope() -> GraphPattern:
    pattern = GraphPattern()
    matmul_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "MATMUL", GraphPattern.METATYPE_ATTR: om.PTMatMulMetatype}
    )
    transpose_node = pattern.add_node(
        **{GraphPattern.LABEL_ATTR: "TRANSPOSE", GraphPattern.METATYPE_ATTR: om.PTTransposeMetatype}
    )
    concat_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "CONCAT", GraphPattern.METATYPE_ATTR: om.PTCatMetatype})
    cos_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "COS", GraphPattern.METATYPE_ATTR: om.PTCosMetatype})
    sin_node = pattern.add_node(**{GraphPattern.LABEL_ATTR: "SIN", GraphPattern.METATYPE_ATTR: om.PTSinMetatype})

    pattern.add_edge(matmul_node, transpose_node)
    pattern.add_edge(transpose_node, concat_node)
    pattern.add_edge(concat_node, cos_node)
    pattern.add_edge(concat_node, sin_node)
    return pattern



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import hashlib
from enum import Enum
from typing import Any, Dict, Tuple

import networkx as nx  # type: ignore[import-untyped]
import pydot  # type: ignore[import-untyped]

from nncf.experimental.torch2.function_hook.graph.graph_utils import ConstMeta
from nncf.experimental.torch2.function_hook.graph.graph_utils import EdgeMeta
from nncf.experimental.torch2.function_hook.graph.graph_utils import FunctionMeta
from nncf.experimental.torch2.function_hook.graph.graph_utils import InOutMeta


class PydotStyleTemplate(Enum):
    """
    Enum to define different styles for Pydot graph representation.
     - disable: labels contain only names, used for tests  (not recommend to convert to svg)
     - short: labels contain names, add edge info
     - full: labels contain all full information about nodes
    """

    disable = "disable"
    short = "short"
    full = "full"

    def __str__(self) -> str:
        return self.value


def fix_dot_label(label: str) -> str:
    """
    Escapes curly braces in a DOT label to avoid syntax errors.

    :param label: The label string to be fixed.
    :return: The label with escaped curly braces.
    """
    return label.replace("{", r"\{").replace("}", r"\}")


def args_to_label(args: Tuple[Any, ...]) -> str:
    """
    Converts function arguments to a formatted string label.

    :param args: Function arguments.
    :return: Formatted string label of arguments.
    """
    if not args:
        return "[]"
    ret = "["
    for arg in args:
        ret += f"\n{arg},"
    return ret + "\n]"


def kwargs_to_label(kwargs: Dict[str, Any]) -> str:
    """
    Converts function keyword arguments to a formatted string label.

    :param kwargs: Function keyword arguments.
    :return: Formatted string label of keyword arguments.
    """
    if not kwargs:
        return "{}"
    ret = "{"
    for key, val in kwargs.items():
        ret += f"\n{key} : {str(val)[:50]}"
    return ret + "\n}"


def get_label_from_node_data(node_data: Dict[str, Any], style: PydotStyleTemplate) -> str:
    """
    Generates a label for a graph node based on its metadata and the desired style.

    :param node_data: Metadata of the node.
    :param style: Style template to determine the label format.
    :return: Formatted label for the node.
    """
    meta = node_data["meta"]
    node_type = node_data["type"]
    if style == PydotStyleTemplate.full:
        rows = []
        if isinstance(meta, InOutMeta):
            rows = [
                f"type: {node_type}",
                f"name: {meta.name}",
                f"dtype: {meta.dtype}",
                f"shape: {meta.shape}",
            ]
        elif isinstance(meta, ConstMeta):
            rows = [
                f"type: {node_type}",
                f"name: {meta.name_in_model}",
                f"dtype: {meta.dtype}",
                f"shape: {meta.shape}",
            ]
        if isinstance(meta, FunctionMeta):
            rows = [
                f"type: {node_type}",
                f"op_name: {meta.op_name}",
                f"fn_name: {meta.fn_name}",
                f"args: {args_to_label(meta.args)}",
                f"kwargs: {kwargs_to_label(meta.kwargs)}",
            ]
        return "{" + fix_dot_label("|".join(rows)) + "}"
    else:
        if isinstance(meta, InOutMeta):
            return f"{meta.name}"
        if isinstance(meta, ConstMeta):
            return f"{meta.name_in_model}"
        if isinstance(meta, FunctionMeta):
            return f"{meta.op_name}"
    raise ValueError(f"Unknown meta node {type(meta)}")


def get_label_from_edge_data(node_data: Dict[str, Any], style: PydotStyleTemplate) -> str:
    """
    Generates a label for a graph edge based on its metadata and the desired style.

    :param edge_data: Metadata of the edge.
    :param style: Style template to determine the label format.
    :return: Formatted label for the edge.
    """
    meta = node_data["meta"]
    assert isinstance(meta, EdgeMeta)

    if style == PydotStyleTemplate.disable:
        return f"{meta.output_port} → {meta.input_port}"
    else:
        return f"{meta.shape}\n{meta.output_port} → {meta.input_port}"


_RAINBOW_COLORS = [
    "#ffadad",
    "#ffc2a9",
    "#ffd6a5",
    "#fdffb6",
    "#caffbf",
    "#b3fbdf",
    "#aae0ef",
    "#a0c4ff",
    "#bdb2ff",
    "#ffc6ff",
]


def color_picker(data: str) -> str:
    """
    Picks a color from a predefined set of colors based on the input string.

    :param data: Input string to determine the color.
    :return: Hex code of the selected color.
    """
    data = "".join(d for d in data if d.isalpha())
    hash_object = hashlib.sha256(data.encode())
    hex_int = int(hash_object.hexdigest()[:6], 16)
    return _RAINBOW_COLORS[hex_int % len(_RAINBOW_COLORS)]


def get_style(node: Dict[str, Any], style: PydotStyleTemplate) -> Dict[str, str]:
    """
    Generates a style dictionary for a graph node based on its metadata and the desired style.

    :param node: Metadata of the node.
    :param style: Style template to determine the node style.
    :return: Dictionary containing style attributes for the node.
    """
    if style == PydotStyleTemplate.disable:
        return {}
    meta = node["meta"]

    if isinstance(meta, InOutMeta):
        return {
            "fillcolor": "#adadad",
            "fontcolor": "#000000",
            "shape": "record",
            "style": '"filled,rounded"',
        }
    if isinstance(meta, ConstMeta):
        return {
            "fillcolor": "#ffffff",
            "fontcolor": "#000000",
            "shape": "record",
            "style": '"filled,rounded"',
        }
    if isinstance(meta, FunctionMeta):
        return {
            "fillcolor": color_picker(meta.fn_name),
            "fontcolor": "#000000",
            "shape": "record",
            "style": '"filled,rounded"',
        }

    raise ValueError(f"Unknown meta node {type(meta)}")


def to_pydot(nx_graph: nx.MultiDiGraph, style_template: PydotStyleTemplate = PydotStyleTemplate.full) -> pydot.Graph:
    """
    Converts a NetworkX directed graph to a Pydot graph with specified styling.

    :param nx_graph: Input NetworkX directed graph.
    :param style_template: Style template to determine node and edge styles.
    :return: Pydot graph representation of the input NetworkX graph.
    """
    dot_graph = pydot.Dot("", rankdir="TB")

    for key, data in nx_graph.nodes(data=True):
        style = get_style(data, style_template)
        dot_node = pydot.Node(key, label=get_label_from_node_data(data, style_template), **style)
        dot_graph.add_node(dot_node)

    for key_from, key_to, data in nx_graph.edges(data=True):
        dot_edge = pydot.Edge(key_from, key_to, label=get_label_from_edge_data(data, style_template))
        dot_graph.add_edge(dot_edge)

    return dot_graph



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import gc
import weakref

import tensorflow as tf

from nncf.config.utils import is_experimental_quantization


class TFOriginalModelManager:
    def __init__(self, model_fn, *args, **kwargs):
        self._model_fn = model_fn
        self._kwargs = kwargs
        self._args = args
        self._model = None

    def __enter__(self):
        self._model = self._model_fn(*self._args, **self._kwargs)
        tf.keras.backend.clear_session()
        return weakref.proxy(self._model)

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self._model
        gc.collect()


class TFWithoutModelManager(TFOriginalModelManager):
    def __enter__(self):
        return self._model_fn(*self._args, **self._kwargs)

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class TFModelManager:
    """
    Controls the process of model creation.
    """

    def __init__(self, model_fn, nncf_config, *args, **kwargs):
        """
        Initializes the `TFModelManager`.

        :param model_fn: Function for model creation.
        :param nncf_config: NNCF config.
        """
        self._manager = TFModelManager._create_model_manager(model_fn, nncf_config, *args, **kwargs)

    def __enter__(self):
        return self._manager.__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self._manager.__exit__(exc_type, exc_val, exc_tb)

    @staticmethod
    def _create_model_manager(model_fn, nncf_config, *args, **kwargs):
        """
        Creates model manager depending on the algorithm name.

        :param model_fn: Function for model creation.
        :param nncf_config: NNCF config.
        :return: Model manager.
        """
        if is_experimental_quantization(nncf_config):
            return TFWithoutModelManager(model_fn, *args, **kwargs)
        return TFOriginalModelManager(model_fn, *args, **kwargs)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from abc import ABC
from abc import abstractmethod
from typing import Optional, Type

from nncf.api.compression import TModel
from nncf.common.initialization.dataloader import NNCFDataLoader
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend


class BatchnormAdaptationAlgorithmImpl(ABC):
    """
    This is the class from which all framework-specific implementations of
    the batch-norm statistics adaptation algorithm inherit.
    """

    def __init__(self, data_loader: NNCFDataLoader, num_bn_adaptation_steps: int, device: Optional[str] = None):
        """
        Initializes the batch-norm statistics adaptation algorithm implementation.

        :param data_loader: NNCF data loader.
        :param num_bn_adaptation_steps: Number of batches from the training dataset to pass
            through the model at initialization in order to update batch-norm statistics of
            the original model.
        :param device: Device to perform initialization. If `device` is `None` then the device
            of the model parameters will be used.
        """
        self._data_loader = data_loader
        self._num_bn_adaptation_steps = num_bn_adaptation_steps
        self._device = device

    @abstractmethod
    def run(self, model: TModel) -> None:
        """
        Runs the batch-norm statistics adaptation algorithm. This method contains the implementation
        of the algorithm.
        """


class BatchnormAdaptationAlgorithm:
    """
    This algorithm updates the statistics of the batch normalization layers
    passing several batches of data through the model. This allows to correct
    the compression-induced bias in the model and reduce the corresponding
    accuracy drop even before model training.
    """

    def __init__(self, data_loader: NNCFDataLoader, num_bn_adaptation_samples: int, device: Optional[str] = None):
        """
        Initializes the batch-norm statistics adaptation algorithm.

        :param data_loader: NNCF data loader.
        :param num_bn_adaptation_samples: Number of samples from the training
            dataset to pass through the model at initialization in order to update
            batch-norm statistics of the original model. The actual number of samples
            will be a closest multiple of the batch size.
        :param device: Device to perform initialization. If `device` is `None` then the device
            of the model parameters will be used.
        """
        if num_bn_adaptation_samples < 0:
            raise ValueError("Number of adaptation samples must be >= 0")

        self._device = device
        self._data_loader = data_loader
        self._num_bn_adaptation_steps = math.ceil(num_bn_adaptation_samples / data_loader.batch_size)

    def run(self, model: TModel) -> None:
        """
        Runs the batch-norm statistics adaptation algorithm.

        :param model: A model for which the algorithm will be applied.
        """
        backend = get_backend(model)
        impl_cls: Type[BatchnormAdaptationAlgorithmImpl]
        if backend is BackendType.TORCH:
            from nncf.torch.batchnorm_adaptation import PTBatchnormAdaptationAlgorithmImpl

            impl_cls = PTBatchnormAdaptationAlgorithmImpl
        else:
            assert backend is BackendType.TENSORFLOW
            from nncf.tensorflow.batchnorm_adaptation import TFBatchnormAdaptationAlgorithmImpl

            impl_cls = TFBatchnormAdaptationAlgorithmImpl
        impl = impl_cls(self._data_loader, self._num_bn_adaptation_steps, self._device)
        impl.run(model)



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List

from nncf.common.quantization.quantizer_setup import SingleConfigQuantizerSetup
from nncf.torch.quantization.precision_constraints import HardwareQuantizationConstraints
from nncf.torch.quantization.precision_init.base_init import BasePrecisionInitializer
from nncf.torch.quantization.precision_init.base_init import BasePrecisionInitParams
from nncf.torch.structures import QuantizationPrecisionInitArgs


class ManualPrecisionInitParams(BasePrecisionInitParams):
    def __init__(self, user_init_args: QuantizationPrecisionInitArgs = None, bitwidth_per_scope: List[List] = None):
        super().__init__(user_init_args)
        self.bitwidth_per_scope = bitwidth_per_scope

    @classmethod
    def from_config(cls, manual_init_params_dict: Dict):
        return cls(user_init_args=None, bitwidth_per_scope=manual_init_params_dict.get("bitwidth_per_scope", []))


class ManualPrecisionInitializer(BasePrecisionInitializer):
    def __init__(
        self,
        algo: "ExperimentalQuantizationController",  # noqa: F821
        params: ManualPrecisionInitParams,
        hw_precision_constraints: HardwareQuantizationConstraints = None,
    ):
        super().__init__(algo, params, hw_precision_constraints)
        self._bitwidth_per_scope = params.bitwidth_per_scope

    def apply_init(self) -> SingleConfigQuantizerSetup:
        quantizer_setup = self._algo.get_quantizer_setup_for_current_state()
        for pair in self._bitwidth_per_scope:
            bitwidth, scope_name = pair
            is_matched = False
            msg = (
                "Failed to assign bitwidth={} to `{}`,\n"
                "because it is incompatible for the specified target hardware\n"
                "Supported quantization configs: {}"
            )
            for qp_id, qp in quantizer_setup.quantization_points.items():
                if scope_name in str(qp.insertion_point):
                    if self._hw_precision_constraints:
                        q_id = self._algo.setup_to_module_id_translation_dict[qp_id]
                        q_configs = self._hw_precision_constraints.get(q_id)
                        matched_q_configs = list(filter(lambda x: x.num_bits == bitwidth, q_configs))
                        if not matched_q_configs:
                            raise ValueError(msg.format(bitwidth, scope_name, list(map(str, q_configs))))
                        qp.qconfig = matched_q_configs[0]
                    else:
                        qp.qconfig.num_bits = bitwidth
                    is_matched = True
                    break
            if not is_matched:
                raise ValueError(
                    "Could not find a quantization point at scope name `{}`, failed to assign bitwidth {} "
                    "to it".format(scope_name, bitwidth)
                )
        return quantizer_setup



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from typing import Optional, Tuple, Type

import numpy as np

from nncf.experimental.common.tensor_statistics.collectors import AbsMaxReducer
from nncf.experimental.common.tensor_statistics.collectors import AggregatorBase
from nncf.experimental.common.tensor_statistics.collectors import BatchMeanReducer
from nncf.experimental.common.tensor_statistics.collectors import MaxAggregator
from nncf.experimental.common.tensor_statistics.collectors import MaxReducer
from nncf.experimental.common.tensor_statistics.collectors import MeanAggregator
from nncf.experimental.common.tensor_statistics.collectors import MeanPerChReducer
from nncf.experimental.common.tensor_statistics.collectors import MedianAbsoluteDeviationAggregator
from nncf.experimental.common.tensor_statistics.collectors import MinAggregator
from nncf.experimental.common.tensor_statistics.collectors import MinReducer
from nncf.experimental.common.tensor_statistics.collectors import NoopAggregator
from nncf.experimental.common.tensor_statistics.collectors import PercentileAggregator
from nncf.experimental.common.tensor_statistics.collectors import QuantileReducer
from nncf.experimental.common.tensor_statistics.collectors import RawReducer
from nncf.experimental.common.tensor_statistics.collectors import ShapeAggregator
from nncf.experimental.common.tensor_statistics.collectors import TensorCollector
from nncf.experimental.common.tensor_statistics.statistics import MeanTensorStatistic
from nncf.experimental.common.tensor_statistics.statistics import MedianMADTensorStatistic
from nncf.experimental.common.tensor_statistics.statistics import MinMaxTensorStatistic
from nncf.experimental.common.tensor_statistics.statistics import PercentileTensorStatistic
from nncf.experimental.common.tensor_statistics.statistics import RawTensorStatistic
from nncf.tensor import Tensor


def _reshape_all(targets: Tuple[Tensor, ...], target_shape: Tuple[int, ...]):
    return map(lambda stat: stat.reshape(target_shape), targets)


def _get_wrapped_min_max_tensor_statistic(target_shape: Tuple[int, ...]) -> Type[MinMaxTensorStatistic]:
    """
    Returns MinMaxTensorStatistic type but all statistics are reshaped to target_shape.

    :param target_shape: Target shape of the tensor statistic
    :return: MinMaxTensorStatistic type but all statistics are reshaped to target_shape.
    """

    class WrappedPTMinMaxTensorStatistic(MinMaxTensorStatistic):
        def __init__(self, min_values, max_values):
            min_values, max_values = _reshape_all((min_values, max_values), target_shape)
            super().__init__(min_values, max_values)

    return WrappedPTMinMaxTensorStatistic


def _get_wrapped_percentile_tensor_statistic(target_shape: Tuple[int, ...]) -> Type[PercentileTensorStatistic]:
    """
    Returns PercentileTensorStatistic type but all statistics are reshaped to target_shape.

    :param target_shape: Target shape of the tensor statistic
    :return: PercentileTensorStatistic type but all statistics are reshaped to target_shape.
    """

    class WrappedPTPercentileTensorStatistic(PercentileTensorStatistic):
        def __init__(self, percentile_vs_values_dict):
            reshaped_percentiles = {}
            for k, v in percentile_vs_values_dict.items():
                reshaped_percentiles[k] = v.reshape(target_shape)
            super().__init__(reshaped_percentiles)

    return WrappedPTPercentileTensorStatistic


def get_min_max_statistic_collector(
    use_abs_max: bool,
    reduction_axes: Tuple[int, ...],
    aggregation_axes: Tuple[int, ...],
    scale_shape: Tuple[int, ...],
    num_samples: int,
) -> TensorCollector:
    """
    Min max statistic collector builder.

    :param use_abs_max: Whether to use abs max reducer or max reducer.
    :param reduction_axes: Axes to use in reduction functions.
    :param aggregation_axes: Axes to use in aggregation functions.
    :param scale_shape: Target shape for collected statistics.
    :param num_samples: Maximum number of samples to collect.
    :return: Min max statistic collector.
    """

    tensor_collector = TensorCollector(_get_wrapped_min_max_tensor_statistic(target_shape=scale_shape))

    aggregator_kwargs = {
        "num_samples": num_samples,
        "aggregation_axes": aggregation_axes,
    }
    min_reducer = MinReducer(reduction_axes)
    min_aggregator = MinAggregator(**aggregator_kwargs)
    tensor_collector.register_statistic_branch(MinMaxTensorStatistic.MIN_STAT, min_reducer, min_aggregator)

    max_reducer_cls = AbsMaxReducer if use_abs_max else MaxReducer
    max_reducer = max_reducer_cls(reduction_axes)
    max_aggregator = MaxAggregator(**aggregator_kwargs)
    tensor_collector.register_statistic_branch(MinMaxTensorStatistic.MAX_STAT, max_reducer, max_aggregator)
    return tensor_collector


def get_mixed_min_max_statistic_collector(
    use_abs_max: bool,
    reduction_axes: Tuple[int, ...],
    aggregation_axes: Tuple[int, ...],
    scale_shape: Tuple[int, ...],
    use_means_of_mins: bool,
    use_means_of_maxs: bool,
    num_samples: int = None,
    window_size: Optional[int] = None,
) -> TensorCollector:
    """
    Mixed min max statistic collector builder.

    :param use_abs_max: Whether to use abs max reducer or max reducer.
    :param reduction_axes: Axes to use in reduction functions.
    :param aggregation_axes: Axes to use in aggregation functions.
    :param scale_shape: Target shape for collected statistics.
    :param use_means_of_mins: Whether to use mean or min aggregator for minimum statistic branch.
    :param use_means_of_maxs: Whether to use mean or max aggregator for maximum statistic branch.
    :param num_samples: Maximum number of samples to collect.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :return: Mixed min max statistic collector.
    """
    tensor_collector = TensorCollector(_get_wrapped_min_max_tensor_statistic(target_shape=scale_shape))
    min_reducer = MinReducer(reduction_axes)

    kwargs = {
        "num_samples": num_samples,
        "aggregation_axes": aggregation_axes,
        "window_size": window_size,
    }
    min_aggregator_cls = MeanAggregator if use_means_of_mins else MinAggregator
    min_aggregator = min_aggregator_cls(**kwargs)
    tensor_collector.register_statistic_branch(MinMaxTensorStatistic.MIN_STAT, min_reducer, min_aggregator)

    max_reducer_cls = AbsMaxReducer if use_abs_max else MaxReducer
    max_reducer = max_reducer_cls(reduction_axes)
    max_aggregator_cls = MeanAggregator if use_means_of_maxs else MaxAggregator
    max_aggregator = max_aggregator_cls(**kwargs)
    tensor_collector.register_statistic_branch(MinMaxTensorStatistic.MAX_STAT, max_reducer, max_aggregator)

    return tensor_collector


def get_median_mad_statistic_collector(
    reduction_axes: Tuple[int, ...],
    aggregation_axes: Tuple[int, ...],
    scale_shape: Tuple[int, ...],
    num_samples: int,
    window_size: Optional[int] = None,
) -> TensorCollector:
    """
    Median Absolute Deviation statistic collector builder.

    :param reduction_axes: Axes to use in reduction functions.
    :param aggregation_axes: Axes to use in aggregation functions.
    :param scale_shape: Target shape for collected statistics.
    :param num_samples: Maximum number of samples to collect.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :return: Median Absolute Deviation statistic collector.

    """

    class WrappedPTMedianMADTensorStatistic(MedianMADTensorStatistic):
        def __init__(self, median_values, mad_values):
            median_values, mad_values = _reshape_all((median_values, mad_values), scale_shape)
            super().__init__(median_values, mad_values)

    return _get_collection_without_reduction(
        MedianAbsoluteDeviationAggregator,
        WrappedPTMedianMADTensorStatistic,
        reduction_axes=reduction_axes,
        aggregation_axes=aggregation_axes,
        num_samples=num_samples,
        window_size=window_size,
    )


def get_percentile_tensor_collector(
    percentiles_to_collect: Tuple[int, ...],
    reduction_axes: Tuple[int, ...],
    aggregation_axes: Tuple[int, ...],
    scale_shape: Tuple[int, ...],
    num_samples: int,
    window_size: Optional[int] = None,
) -> TensorCollector:
    """
    Percentile statistic collector builder.

    :param percentiles_to_collect: Percentiles to use on aggregation phase.
    :param reduction_axes: Axes to use in reduction functions.
    :param aggregation_axes: Axes to use in aggregation functions.
    :param scale_shape: Target shape for collected statistics.
    :param num_samples: Maximum number of samples to collect.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :return: Percentile statistic collector.
    """
    return _get_collection_without_reduction(
        partial(PercentileAggregator, percentiles_to_collect=percentiles_to_collect),
        _get_wrapped_percentile_tensor_statistic(target_shape=scale_shape),
        reduction_axes=reduction_axes,
        aggregation_axes=aggregation_axes,
        num_samples=num_samples,
        window_size=window_size,
    )


def _get_collection_without_reduction(
    aggregator_cls: AggregatorBase,
    statistic_cls: AggregatorBase,
    reduction_axes: Tuple[int, ...],
    aggregation_axes: Tuple[int, ...],
    num_samples: int,
    window_size: Optional[int] = None,
) -> TensorCollector:
    """
    Helper function to build a tensor collector which is reducing statistics exclusively during aggregation phase.

    :param aggregator_cls: Aggregator class to build the tensor collector.
    :param aggregator_cls: Statistic class to build the tensor collector.
    :param reduction_axes: Axes to use in reduction functions.
    :param aggregation_axes: Axes to use in aggregation functions.
    :param num_samples: Maximum number of samples to collect.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :return: Target statistic collector.
    """
    tensor_collector = TensorCollector(statistic_cls)
    reducer = RawReducer()
    aggregation_axes = list(set(list(aggregation_axes) + [dim + 1 for dim in reduction_axes]))
    aggregator = aggregator_cls(
        aggregation_axes=aggregation_axes,
        window_size=window_size,
        num_samples=num_samples,
    )

    tensor_collector.register_statistic_branch(
        MedianMADTensorStatistic.TENSOR_STATISTIC_OUTPUT_KEY, reducer, aggregator
    )
    return tensor_collector


def get_mean_percentile_statistic_collector(
    percentiles_to_collect: Tuple[int, ...],
    reduction_axes: Tuple[int, ...],
    aggregation_axes: Tuple[int, ...],
    scale_shape: Tuple[int, ...],
    num_samples: int,
    window_size: Optional[int] = None,
) -> TensorCollector:
    """
    Mean percentile statistic collector builder.

    :param percentiles_to_collect: Percentiles to use on reduction phase.
    :param reduction_axes: Axes to use in reduction functions.
    :param aggregation_axes: Axes to use in aggregation functions.
    :param scale_shape: Target shape for collected statistics.
    :param num_samples: Maximum number of samples to collect.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :return: Mean percentile statistic collector.
    """
    tensor_collector = TensorCollector(_get_wrapped_percentile_tensor_statistic(target_shape=scale_shape))
    quantiles_to_collect = np.true_divide(percentiles_to_collect, 100)
    reducer = QuantileReducer(reduction_axes=reduction_axes, quantile=quantiles_to_collect)
    for output_port_id, p in enumerate(percentiles_to_collect):
        aggregator = MeanAggregator(
            aggregation_axes=aggregation_axes,
            num_samples=num_samples,
            window_size=window_size,
        )
        tensor_collector.register_statistic_branch(
            (PercentileTensorStatistic.PERCENTILE_VS_VALUE_DICT, p), reducer, aggregator, output_port_id
        )
    return tensor_collector


def get_mean_statistic_collector(
    num_samples: int, channel_axis: int, window_size: Optional[int] = None
) -> TensorCollector:
    """
    Mean statistic collector builder.

    :param num_samples: Maximum number of samples to collect.
    :param channel_axis: Channel axis to use during reduction phase.
    :param window_size: Number of samples from the end of the list of collected samples to aggregate.
        Aggregates all available collected statistics in case parameter is None.
    :return: Mean statistic collector.
    """
    if channel_axis == 0:
        reducer = BatchMeanReducer()
    else:
        reducer = MeanPerChReducer(channel_axis=channel_axis)
    raw_reducer = RawReducer()

    kwargs = {
        "num_samples": num_samples,
        "window_size": window_size,
    }
    aggregate_mean = MeanAggregator(**kwargs)
    aggregate_shape = ShapeAggregator()

    collector = TensorCollector(MeanTensorStatistic)
    collector.register_statistic_branch(MeanTensorStatistic.MEAN_STAT, reducer, aggregate_mean)
    collector.register_statistic_branch(MeanTensorStatistic.SHAPE_STAT, raw_reducer, aggregate_shape)
    return collector


def get_raw_stat_collector(num_samples: Optional[int] = None) -> TensorCollector:
    """
    Raw statistic collector builder.

    :param num_samples: Maximum number of samples to collect.
    :return: Raw statistic collector.
    """
    reducer = RawReducer()
    aggregator = NoopAggregator(num_samples)

    collector = TensorCollector(RawTensorStatistic)
    collector.register_statistic_branch(RawTensorStatistic.VALUES_STATS, reducer, aggregator)
    return collector



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from dataclasses import dataclass
from dataclasses import fields
from typing import Iterator, Optional, Tuple

from nncf.api.statistics import Statistics
from nncf.common.pruning.statistics import FilterPruningStatistics
from nncf.common.quantization.statistics import QuantizationStatistics
from nncf.common.sparsity.statistics import ConstSparsityStatistics
from nncf.common.sparsity.statistics import MagnitudeSparsityStatistics
from nncf.common.sparsity.statistics import MovementSparsityStatistics
from nncf.common.sparsity.statistics import RBSparsityStatistics
from nncf.common.utils.api_marker import api


@api()
@dataclass
class NNCFStatistics:
    """
    Groups statistics for all available NNCF compression algorithms.
    Statistics are present only if the algorithm has been started.
    """

    const_sparsity: Optional[ConstSparsityStatistics] = None
    filter_pruning: Optional[FilterPruningStatistics] = None
    magnitude_sparsity: Optional[MagnitudeSparsityStatistics] = None
    movement_sparsity: Optional[MovementSparsityStatistics] = None
    quantization: Optional[QuantizationStatistics] = None
    rb_sparsity: Optional[RBSparsityStatistics] = None

    def register(self, algorithm_name: str, stats: Statistics) -> None:
        """
        Registers statistics for the algorithm.

        :param algorithm_name: Name of the algorithm. Should be one of the following
            * const_sparsity
            * filter_pruning
            * magnitude_sparsity
            * movement_sparsity
            * quantization
            * rb_sparsity

        :param stats: Statistics of the algorithm.
        """
        available_algorithms = [f.name for f in fields(self)]
        if algorithm_name not in available_algorithms:
            raise ValueError(
                f"Can not register statistics for the algorithm. Unknown name of the algorithm: {algorithm_name}."
            )

        setattr(self, algorithm_name, stats)

    def to_str(self) -> str:
        pretty_string = "\n\n".join([x[1].to_str() for x in self])
        return pretty_string

    def __iter__(self) -> Iterator[Tuple[str, Statistics]]:
        return iter([(f.name, getattr(self, f.name)) for f in fields(self) if getattr(self, f.name) is not None])



# Copyright (c) 2025 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List

from nncf.api.statistics import Statistics
from nncf.common.utils.api_marker import api
from nncf.common.utils.helpers import create_table


class PrunedLayerSummary:
    """
    Contains information about the pruned layer.
    """

    def __init__(self, name: str, weight_shape: List[int], mask_shape: List[int], filter_pruning_level: float):
        """
        Initializes a summary about the pruned layer.

        :param name: Layer's name.
        :param weight_shape: Weight's shape.
        :param mask_shape: Mask's shape.
        :param filter_pruning_level: Filter's pruning level.
        """
        self.name = name
        self.weight_shape = weight_shape
        self.mask_shape = mask_shape
        self.filter_pruning_level = filter_pruning_level


class PrunedModelStatistics(Statistics):
    """
    Contains statistics of the pruned model.
    """

    def __init__(
        self,
        full_flops: int,
        current_flops: int,
        full_params_num: int,
        current_params_num: int,
        full_filters_num: int,
        current_filters_num: int,
        pruned_layers_summary: List[PrunedLayerSummary],
    ):
        """
        Initializes statistics of the pruned model.

        :param full_flops: The total amount of FLOPs in the model.
        :param current_flops: Current amount of FLOPs in the model.
        :param full_params_num: The total amount of weights in the model.
        :param current_params_num: Current amount of weights in the model.
        :param full_filters_num: The total amount of filters in the model.
        :param current_filters_num: Current amount of filters in the model.
        :param pruned_layers_summary: Detailed summary for the
            pruned layers.
        """
        self._giga = 1e9
        self._mega = 1e6
        self.full_flops = full_flops
        self.current_flops = current_flops
        self.flops_pruning_level = 1 - self.current_flops / self.full_flops
        self.full_params_num = full_params_num
        self.current_params_num = current_params_num
        self.params_pruning_level = 1 - self.current_params_num / self.full_params_num
        self.full_filters_num = full_filters_num
        self.current_filters_num = current_filters_num
        self.filter_pruning_level = 1 - self.current_filters_num / self.full_filters_num
        self.pruned_layers_summary = pruned_layers_summary

    def to_str(self) -> str:
        model_string = create_table(
            header=["#", "Full", "Current", "Pruning level"],
            rows=[
                [
                    "GFLOPs",
                    f"{self.full_flops / self._giga:.3f}",
                    f"{self.current_flops / self._giga:.3f}",
                    self.flops_pruning_level,
                ],
                [
                    "MParams",
                    f"{self.full_params_num / self._mega:.3f}",
                    f"{self.current_params_num / self._mega:.3f}",
                    self.params_pruning_level,
                ],
                ["Filters", self.full_filters_num, self.current_filters_num, self.filter_pruning_level],
            ],
        )

        header = ["Layer's name", "Weight's shape", "Mask's shape", "Filter pruning level"]
        rows = []
        for s in self.pruned_layers_summary:
            rows.append([s.name, s.weight_shape, s.mask_shape, s.filter_pruning_level])

        layers_string = create_table(header, rows)

        pruning_level_desc = "Prompt: statistic pruning level = 1 - statistic current / statistic full."
        pretty_string = (
            f"Statistics by pruned layers:\n{layers_string}\n"
            f"Statistics of the pruned model:\n{model_string}\n" + pruning_level_desc
        )

        return pretty_string


@api()
class FilterPruningStatistics(Statistics):
    """
    Contains statistics of the filter pruning algorithm.

    :param model_statistics: Statistics of the pruned model.
    :param current_pruning_level: A current level of the pruning for the algorithm for the current epoch.
    :param target_pruning_level: A target level of the pruning for the algorithm.
    :param prune_flops: Is pruning algo sets flops pruning level or not (filter pruning level).
    """

    def __init__(
        self,
        model_statistics: PrunedModelStatistics,
        current_pruning_level: float,
        target_pruning_level: float,
        prune_flops: bool,
    ):
        self.model_statistics = model_statistics
        self.current_pruning_level = current_pruning_level
        self.target_pruning_level = target_pruning_level
        self.prune_flops = prune_flops

    def to_str(self) -> str:
        pruning_mode = "FLOPs" if self.prune_flops else "filter"
        algorithm_string = create_table(
            header=["Statistic's name", "Value"],
            rows=[
                [f"{pruning_mode.capitalize()} pruning level in current epoch", self.current_pruning_level],
                [f"Target {pruning_mode} pruning level", self.target_pruning_level],
            ],
        )

        pretty_string = (
            f"{self.model_statistics.to_str()}\nStatistics of the filter pruning algorithm:\n{algorithm_string}"
        )
        return pretty_string


class PrunedModelTheoreticalBorderline(Statistics):
    """
    Contains theoretical borderline statistics of the filter pruning algorithm.
    """

    def __init__(
        self,
        num_pruned_layers: int,
        num_prunable_layers: int,
        min_possible_flops: float,
        min_possible_params: float,
        total_flops: int,
        total_params: int,
    ):
        """
        Initializes statistics of the filter pruning theoretical borderline.

        :param num_pruned_layers: Number of layers which was actually pruned.
        :param num_prunable_layers: Number of layers which have prunable type.
        :param min_possible_flops: Number of flops for pruned model with pruning level = 1.
        :param min_possible_params: Number of weights for pruned model with pruning level = 1.
        :param total_flops: The total amount of FLOPS in the model.
        :param total_params: The total amount of weights in the model.
        """
        self._giga = 1e9
        self._mega = 1e6
        self.pruned_layers_num = num_pruned_layers
        self.prunable_layers_num = num_prunable_layers
        self.min_possible_flops = min_possible_flops
        self.min_possible_params = min_possible_params
        self.total_flops = total_flops
        self.total_params = total_params

    def to_str(self) -> str:
        algorithm_string = create_table(
            header=["Statistic's name", "Value"],
            rows=[
                [
                    "Pruned layers count / prunable layers count",
                    f"{self.pruned_layers_num} / {self.prunable_layers_num}",
                ],
                [
                    "GFLOPs minimum possible after pruning / total",
                    f"{self.min_possible_flops / self._giga:.3f} / {self.total_flops / self._giga:.3f}",
                ],
                [
                    "MParams minimum possible after pruning / total",
                    f"{self.min_possible_params / self._mega:.3f} / {self.total_params / self._mega:.3f}",
                ],
            ],
        )

        pretty_string = (
            f"Theoretical borderline of the filter pruning algorithm\nfor current model:\n{algorithm_string}"
        )
        return pretty_string
